# -*- coding: utf-8 -*-
# Created on 2019-03-19 by hbldh <henrik.blidh@nedomkull.com>
"""
Interface class for the Bleak representation of a GATT Characteristic
"""
from __future__ import annotations

import enum
from collections.abc import Callable
from typing import TYPE_CHECKING, Any, Union
from uuid import UUID

from bleak.assigned_numbers import CharacteristicPropertyName
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.uuids import normalize_uuid_str, uuidstr_to_str

# to prevent circular import
if TYPE_CHECKING:
    from bleak.backends.service import BleakGATTService


class GattCharacteristicsFlags(enum.Enum):
    broadcast = 0x0001
    read = 0x0002
    write_without_response = 0x0004
    write = 0x0008
    notify = 0x0010
    indicate = 0x0020
    authenticated_signed_writes = 0x0040
    extended_properties = 0x0080
    reliable_write = 0x0100
    writable_auxiliaries = 0x0200


class BleakGATTCharacteristic:
    """The Bleak representation of a GATT Characteristic"""

    def __init__(
        self,
        obj: Any,
        handle: int,
        uuid: str,
        properties: list[CharacteristicPropertyName],
        max_write_without_response_size: Callable[[], int],
        service: BleakGATTService,
    ):
        """
        Args:
            obj:
                A platform-specific object for this characteristic.
            max_write_without_response_size:
                The maximum size in bytes that can be written to the
                characteristic in a single write without response command.
            service:
                The service this characteristic belongs to.
        """
        self.obj = obj
        self._handle = handle
        self._uuid = uuid
        self._properties = properties
        self._max_write_without_response_size = max_write_without_response_size
        self._service = service
        self._descriptors: dict[int, BleakGATTDescriptor] = {}

    def __str__(self):
        return f"{self.uuid} (Handle: {self.handle}): {self.description}"

    @property
    def service_uuid(self) -> str:
        """The UUID of the Service containing this characteristic"""
        return self._service.uuid

    @property
    def service_handle(self) -> int:
        """The integer handle of the Service containing this characteristic"""
        return self._service.handle

    @property
    def handle(self) -> int:
        """The handle for this characteristic"""
        return self._handle

    @property
    def uuid(self) -> str:
        """The UUID for this characteristic"""
        return self._uuid

    @property
    def description(self) -> str:
        """Description for this characteristic"""
        return uuidstr_to_str(self.uuid)

    @property
    def properties(self) -> list[CharacteristicPropertyName]:
        """Properties of this characteristic"""
        return self._properties

    @property
    def max_write_without_response_size(self) -> int:
        """
        Gets the maximum size in bytes that can be used for the *data* argument
        of :meth:`BleakClient.write_gatt_char()` when ``response=False``.

        In rare cases, a device may take a long time to update this value, so
        reading this property may return the default value of ``20`` and reading
        it again after a some time may return the expected higher value.

        If you *really* need to wait for a higher value, you can do something
        like this:

        .. code-block:: python

            async with asyncio.timeout(10):
                while char.max_write_without_response_size == 20:
                    await asyncio.sleep(0.5)

        .. warning:: Linux quirk: For BlueZ versions < 5.62, this property
            will always return ``20``.

        .. versionadded:: 0.16
        """

        # for backwards compatibility
        if isinstance(self._max_write_without_response_size, int):
            return self._max_write_without_response_size

        return self._max_write_without_response_size()

    @property
    def descriptors(self) -> list[BleakGATTDescriptor]:
        """List of descriptors for this service"""
        return list(self._descriptors.values())

    def get_descriptor(
        self, specifier: Union[int, str, UUID]
    ) -> Union[BleakGATTDescriptor, None]:
        """Get a descriptor by handle (int) or UUID (str or uuid.UUID)"""
        if isinstance(specifier, int):
            return self._descriptors.get(specifier)

        uuid = normalize_uuid_str(str(specifier))
        for descriptor in self._descriptors.values():
            if descriptor.uuid == uuid:
                return descriptor

        return None

    def add_descriptor(self, descriptor: BleakGATTDescriptor) -> None:
        """Add a :py:class:`~BleakGATTDescriptor` to the characteristic.

        Should not be used by end user, but rather by `bleak` itself.
        """
        if descriptor.handle in self._descriptors:
            raise ValueError(
                f"Descriptor with handle {descriptor.handle} already exists"
            )

        self._descriptors[descriptor.handle] = descriptor



# -*- coding: utf-8 -*-
"""
BLE Client for BlueZ on Linux
"""
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This backend is only available on Linux"

import asyncio
import logging
import os
import warnings
from collections.abc import Callable
from contextlib import AsyncExitStack
from typing import Any, Optional, Union

if sys.version_info < (3, 12):
    from typing_extensions import Buffer, override
else:
    from collections.abc import Buffer
    from typing import override

if sys.version_info < (3, 11):
    from async_timeout import timeout as async_timeout
else:
    from asyncio import timeout as async_timeout

from dbus_fast.aio import MessageBus
from dbus_fast.constants import BusType, ErrorType, MessageType
from dbus_fast.message import Message
from dbus_fast.signature import Variant

from bleak import BleakScanner
from bleak.backends.bluezdbus import defs
from bleak.backends.bluezdbus.manager import get_global_bluez_manager
from bleak.backends.bluezdbus.scanner import BleakScannerBlueZDBus
from bleak.backends.bluezdbus.utils import assert_reply, get_dbus_authenticator
from bleak.backends.bluezdbus.version import BlueZFeatures
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.client import BaseBleakClient, NotifyCallback
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTServiceCollection
from bleak.exc import BleakDBusError, BleakDeviceNotFoundError, BleakError

logger = logging.getLogger(__name__)

# prevent tasks from being garbage collected
_background_tasks: set[asyncio.Task[None]] = set()


class BleakClientBlueZDBus(BaseBleakClient):
    """A native Linux Bleak Client

    Implemented by using the `BlueZ DBUS API <https://docs.ubuntu.com/core/en/stacks/bluetooth/bluez/docs/reference/dbus-api>`_.

    Args:
        address_or_ble_device (`BLEDevice` or str): The Bluetooth address of the BLE peripheral to connect to or the `BLEDevice` object representing it.
        services: Optional list of service UUIDs that will be used.

    Keyword Args:
        timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.
        disconnected_callback (callable): Callback that will be scheduled in the
            event loop when the client is disconnected. The callable must take one
            argument, which will be this client object.
        adapter (str): Bluetooth adapter to use for discovery.
    """

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        services: Optional[set[str]] = None,
        **kwargs: Any,
    ):
        super(BleakClientBlueZDBus, self).__init__(address_or_ble_device, **kwargs)
        # kwarg "device" is for backwards compatibility
        self._adapter: Optional[str] = kwargs.get("adapter", kwargs.get("device"))

        self._device_info: Optional[dict[str, Any]]

        # Backend specific, D-Bus objects and data
        if isinstance(address_or_ble_device, BLEDevice):
            self._device_path = address_or_ble_device.details["path"]
            self._device_info = address_or_ble_device.details.get("props")
        else:
            self._device_path = None
            self._device_info = None

        self._requested_services = services

        # D-Bus message bus
        self._bus: Optional[MessageBus] = None
        # tracks device watcher subscription
        self._remove_device_watcher: Optional[Callable[[], None]] = None
        # private backing for is_connected property
        self._is_connected = False
        # indicates disconnect request in progress when not None
        self._disconnecting_event: Optional[asyncio.Event] = None
        # used to ensure device gets disconnected if event loop crashes
        self._disconnect_monitor_event: Optional[asyncio.Event] = None
        # map of characteristic D-Bus object path to notification callback
        self._notification_callbacks: dict[str, NotifyCallback] = {}
        # map of characteristic D-Bus path to AcquireNotify file descriptor
        self._notification_fds: dict[str, int] = {}

        # used to override mtu_size property
        self._mtu_size: Optional[int] = None

    # Connectivity methods

    @override
    async def connect(
        self, pair: bool, dangerous_use_bleak_cache: bool = False, **kwargs: Any
    ) -> None:
        """Connect to the specified GATT server.

        Keyword Args:
            timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.

        Raises:
            BleakError: If the device is already connected or if the device could not be found.
            BleakDBusError: If there was a D-Bus error
            asyncio.TimeoutError: If the connection timed out
        """
        logger.debug("Connecting to device @ %s", self.address)

        if self.is_connected:
            raise BleakError("Client is already connected")

        if not BlueZFeatures.checked_bluez_version:
            await BlueZFeatures.check_bluez_version()
        if not BlueZFeatures.supported_version:
            raise BleakError("Bleak requires BlueZ >= 5.55.")
        # A Discover must have been run before connecting to any devices.
        # Find the desired device before trying to connect.
        timeout = kwargs.get("timeout", self._timeout)
        if self._device_path is None:
            device = await BleakScanner.find_device_by_address(
                self.address,
                timeout=timeout,
                adapter=self._adapter,
                backend=BleakScannerBlueZDBus,
            )

            if device:
                self._device_info = device.details.get("props")
                self._device_path = device.details["path"]
            else:
                raise BleakDeviceNotFoundError(
                    self.address, f"Device with address {self.address} was not found."
                )

        manager = await get_global_bluez_manager()

        async with async_timeout(timeout):
            while True:
                async with AsyncExitStack() as stack:
                    # Each BLE connection session needs a new D-Bus connection to avoid a
                    # BlueZ quirk where notifications are automatically enabled on reconnect.
                    self._bus = await MessageBus(
                        bus_type=BusType.SYSTEM,
                        negotiate_unix_fd=True,
                        auth=get_dbus_authenticator(),
                    ).connect()

                    stack.callback(self._cleanup_all)

                    def on_connected_changed(connected: bool) -> None:
                        if not connected:
                            logger.debug("Device disconnected (%s)", self._device_path)

                            self._is_connected = False

                            if self._disconnect_monitor_event:
                                self._disconnect_monitor_event.set()
                                self._disconnect_monitor_event = None

                            self._cleanup_all()
                            if self._disconnected_callback is not None:
                                self._disconnected_callback()
                            disconnecting_event = self._disconnecting_event
                            if disconnecting_event:
                                disconnecting_event.set()

                    def on_value_changed(char_path: str, value: bytes) -> None:
                        callback = self._notification_callbacks.get(char_path)

                        if callback:
                            callback(bytearray(value))

                    watcher = manager.add_device_watcher(
                        self._device_path, on_connected_changed, on_value_changed
                    )
                    self._remove_device_watcher = lambda: manager.remove_device_watcher(
                        watcher
                    )

                    self._disconnect_monitor_event = local_disconnect_monitor_event = (
                        asyncio.Event()
                    )

                    # this effectively cancels the disconnect monitor in case the event
                    # was not triggered by a D-Bus callback
                    stack.callback(local_disconnect_monitor_event.set)

                    async def disconnect_device() -> None:
                        # Calling Disconnect cancels any pending connect request. Also,
                        # if connection was successful but _get_services() raises (e.g.
                        # because task was cancelled), then we still need to disconnect
                        # before passing on the exception.
                        if self._bus:
                            # If disconnected callback already fired, this will be a no-op
                            # since self._bus will be None and the _cleanup_all call will
                            # have already disconnected.
                            try:
                                disconnect_reply = await self._bus.call(
                                    Message(
                                        destination=defs.BLUEZ_SERVICE,
                                        interface=defs.DEVICE_INTERFACE,
                                        path=self._device_path,
                                        member="Disconnect",
                                    )
                                )
                                assert disconnect_reply
                                try:
                                    assert_reply(disconnect_reply)
                                except BleakDBusError as e:
                                    # if the object no longer exists, then we know we
                                    # are disconnected for sure, so don't need to log a
                                    # warning about it
                                    if e.dbus_error != ErrorType.UNKNOWN_OBJECT.value:
                                        raise
                            except Exception as e:
                                logger.warning(
                                    f"Failed to cancel connection ({self._device_path}): {e}"
                                )

                    stack.push_async_callback(disconnect_device)

                    # The BlueZ backend does not disconnect devices when the
                    # application closes or crashes. This can cause problems
                    # when trying to reconnect to the same device. To work
                    # around this, we check if the device is already connected.
                    #
                    # For additional details see https://github.com/bluez/bluez/issues/89
                    if manager.is_connected(self._device_path):
                        logger.debug(
                            'skipping calling "Connect" since %s is already connected',
                            self._device_path,
                        )
                    else:
                        logger.debug("Connecting to BlueZ path %s", self._device_path)

                        # Calling pair will fail if we are already paired, so
                        # in that case we just call Connect.
                        if pair and not manager.is_paired(self._device_path):
                            # Trust means device is authorized
                            reply = await self._bus.call(
                                Message(
                                    destination=defs.BLUEZ_SERVICE,
                                    path=self._device_path,
                                    interface=defs.PROPERTIES_INTERFACE,
                                    member="Set",
                                    signature="ssv",
                                    body=[
                                        defs.DEVICE_INTERFACE,
                                        "Trusted",
                                        Variant("b", True),
                                    ],
                                )
                            )
                            assert_reply(reply)

                            # REVIST: This leaves "Trusted" property set if we
                            # fail later. Probably not a big deal since we were
                            # going to trust it anyway.

                            # Pairing means device is authenticated
                            reply = await self._bus.call(
                                Message(
                                    destination=defs.BLUEZ_SERVICE,
                                    interface=defs.DEVICE_INTERFACE,
                                    path=self._device_path,
                                    member="Pair",
                                )
                            )

                            # For resolvable private addresses, the address will
                            # change after pairing, so we need to update that.
                            # Hopefully there is no race condition here. D-Bus
                            # traffic capture shows that Address change happens
                            # at the same time as Paired property change and
                            # that PropertiesChanged signal is sent before the
                            # "Pair" reply is sent.
                            self.address = manager.get_device_address(self._device_path)
                        else:
                            reply = await self._bus.call(
                                Message(
                                    destination=defs.BLUEZ_SERVICE,
                                    interface=defs.DEVICE_INTERFACE,
                                    path=self._device_path,
                                    member="Connect",
                                )
                            )

                        if reply.message_type == MessageType.ERROR:
                            # This error is often caused by RF interference
                            # from other Bluetooth or Wi-Fi devices. In many
                            # cases, retrying will connect successfully.
                            # Note: this error was added in BlueZ 6.62.
                            if (
                                reply.error_name == "org.bluez.Error.Failed"
                                and reply.body
                                and reply.body[0] == "le-connection-abort-by-local"
                            ):
                                logger.debug(
                                    "retry due to le-connection-abort-by-local"
                                )

                                # When this error occurs, BlueZ actually
                                # connected so we get "Connected" property changes
                                # that we need to wait for before attempting
                                # to connect again.
                                await local_disconnect_monitor_event.wait()

                                # Jump way back to the `while True:`` to retry.
                                continue

                            if reply.error_name == ErrorType.UNKNOWN_OBJECT.value:
                                raise BleakDeviceNotFoundError(
                                    self.address,
                                    f"Device with address {self.address} was not found. It may have been removed from BlueZ when scanning stopped.",
                                )

                        assert_reply(reply)

                    self._is_connected = True

                    # Create a task that runs until the device is disconnected.
                    task = asyncio.create_task(
                        self._disconnect_monitor(
                            self._bus,
                            self._device_path,
                            local_disconnect_monitor_event,
                        )
                    )
                    _background_tasks.add(task)
                    task.add_done_callback(_background_tasks.discard)

                    # We will try to use the cache if it exists and `dangerous_use_bleak_cache`
                    # is True.
                    await self._get_services(
                        dangerous_use_bleak_cache=dangerous_use_bleak_cache
                    )

                    stack.pop_all()
                    return

    @staticmethod
    async def _disconnect_monitor(
        bus: MessageBus, device_path: str, disconnect_monitor_event: asyncio.Event
    ) -> None:
        # This task runs until the device is disconnected. If the task is
        # cancelled, it probably means that the event loop crashed so we
        # try to disconnected the device. Otherwise BlueZ will keep the device
        # connected even after Python exits. This will only work if the event
        # loop is called with asyncio.run() or otherwise runs pending tasks
        # after the original event loop stops. This will also cause an exception
        # if a run loop is stopped before the device is disconnected since this
        # task will still be running and asyncio complains if a loop with running
        # tasks is stopped.
        try:
            await disconnect_monitor_event.wait()
        except asyncio.CancelledError:
            try:
                # by using send() instead of call(), we ensure that the message
                # gets sent, but we don't wait for a reply, which could take
                # over one second while the device disconnects.
                await bus.send(
                    Message(
                        destination=defs.BLUEZ_SERVICE,
                        path=device_path,
                        interface=defs.DEVICE_INTERFACE,
                        member="Disconnect",
                    )
                )
            except Exception:
                pass

    def _cleanup_all(self) -> None:
        """
        Free all the allocated resource in DBus. Use this method to
        eventually cleanup all otherwise leaked resources.
        """
        logger.debug("_cleanup_all(%s)", self._device_path)

        if self._remove_device_watcher:
            self._remove_device_watcher()
            self._remove_device_watcher = None

        if not self._bus:
            logger.debug("already disconnected (%s)", self._device_path)
            return

        # Try to disconnect the System Bus.
        try:
            self._bus.disconnect()
        except Exception as e:
            logger.error(
                "Attempt to disconnect system bus failed (%s): %s",
                self._device_path,
                e,
            )
        else:
            # Critical to remove the `self._bus` object here to since it was
            # closed above. If not, calls made to it later could lead to
            # a stuck client.
            self._bus = None

            # Reset all stored services.
            self.services = None

    @override
    async def disconnect(self) -> None:
        """Disconnect from the specified GATT server.

        Raises:
            BleakDBusError: If there was a D-Bus error
            asyncio.TimeoutError if the device was not disconnected within 10 seconds
        """
        logger.debug("Disconnecting ({%s})", self._device_path)

        if self._bus is None:
            # No connection exists. Either one hasn't been created or
            # we have already called disconnect and closed the D-Bus
            # connection.
            logger.debug("already disconnected ({%s})", self._device_path)
            return

        if self._disconnecting_event:
            # another call to disconnect() is already in progress
            logger.debug("already in progress ({%s})", self._device_path)
            async with async_timeout(10):
                await self._disconnecting_event.wait()
        elif self.is_connected:
            self._disconnecting_event = asyncio.Event()
            try:
                # Try to disconnect the actual device/peripheral
                reply = await self._bus.call(
                    Message(
                        destination=defs.BLUEZ_SERVICE,
                        path=self._device_path,
                        interface=defs.DEVICE_INTERFACE,
                        member="Disconnect",
                    )
                )
                assert_reply(reply)
                async with async_timeout(10):
                    await self._disconnecting_event.wait()
            finally:
                self._disconnecting_event = None

        # sanity check to make sure _cleanup_all() was triggered by the
        # "PropertiesChanged" signal handler and that it completed successfully
        assert self._bus is None

    @override
    async def pair(self, *args: Any, **kwargs: Any) -> None:
        """Pair with the peripheral.

        You can use ConnectDevice method if you already know the MAC address of the device.
        Else you need to StartDiscovery, Trust, Pair and Connect in sequence.
        """
        assert self._bus
        # See if it is already paired.
        reply = await self._bus.call(
            Message(
                destination=defs.BLUEZ_SERVICE,
                path=self._device_path,
                interface=defs.PROPERTIES_INTERFACE,
                member="Get",
                signature="ss",
                body=[defs.DEVICE_INTERFACE, "Paired"],
            )
        )
        assert_reply(reply)
        if reply.body[0].value:
            logger.debug("BLE device @ %s is already paired", self.address)
            return

        # Set device as trusted.
        reply = await self._bus.call(
            Message(
                destination=defs.BLUEZ_SERVICE,
                path=self._device_path,
                interface=defs.PROPERTIES_INTERFACE,
                member="Set",
                signature="ssv",
                body=[defs.DEVICE_INTERFACE, "Trusted", Variant("b", True)],
            )
        )
        assert_reply(reply)

        logger.debug("Pairing to BLE device @ %s", self.address)

        reply = await self._bus.call(
            Message(
                destination=defs.BLUEZ_SERVICE,
                path=self._device_path,
                interface=defs.DEVICE_INTERFACE,
                member="Pair",
            )
        )
        assert_reply(reply)

        # For resolvable private addresses, the address will
        # change after pairing, so we need to update that.
        # Hopefully there is no race condition here. D-Bus
        # traffic capture shows that Address change happens
        # at the same time as Paired property change and
        # that PropertiesChanged signal is sent before the
        # "Pair" reply is sent.
        manager = await get_global_bluez_manager()
        self.address = manager.get_device_address(self._device_path)

    @override
    async def unpair(self) -> None:
        """Unpair with the peripheral."""
        adapter_path = await self._get_adapter_path()
        device_path = await self._get_device_path()
        manager = await get_global_bluez_manager()

        logger.debug(
            "Removing BlueZ device path %s from adapter path %s",
            device_path,
            adapter_path,
        )

        # If this client object wants to connect again, BlueZ needs the device
        # to follow Discovery process again - so reset the local connection
        # state.
        #
        # (This is true even if the request to RemoveDevice fails,
        # so clear it before.)
        self._device_path = None
        self._device_info = None
        self._is_connected = False

        assert manager._bus

        try:
            reply = await manager._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=adapter_path,
                    interface=defs.ADAPTER_INTERFACE,
                    member="RemoveDevice",
                    signature="o",
                    body=[device_path],
                )
            )
            assert_reply(reply)
        except BleakDBusError as e:
            if e.dbus_error == "org.bluez.Error.DoesNotExist":
                raise BleakDeviceNotFoundError(
                    self.address, f"Device with address {self.address} was not found."
                ) from e
            raise

    @property
    @override
    def is_connected(self) -> bool:
        """Check connection status between this client and the server.

        Returns:
            Boolean representing connection status.

        """
        return False if self._bus is None else self._is_connected

    async def _acquire_mtu(self) -> None:
        """Acquires the MTU for this device by calling the "AcquireWrite" or
        "AcquireNotify" method of the first characteristic that has such a method.

        This method only needs to be called once, after connecting to the device
        but before accessing the ``mtu_size`` property.

        If a device uses encryption on characteristics, it will need to be bonded
        first before calling this method.
        """

        assert (
            self.services is not None
        ), "Services must be acquired before acquiring MTU"

        # This will try to get the "best" characteristic for getting the MTU.
        # We would rather not start notifications if we don't have to.
        try:
            method = "AcquireWrite"
            char = next(
                c
                for c in self.services.characteristics.values()
                if "write-without-response" in c.properties
            )
        except StopIteration:
            method = "AcquireNotify"
            char = next(
                c
                for c in self.services.characteristics.values()
                if "notify" in c.properties
            )

        assert self._bus

        reply = await self._bus.call(
            Message(
                destination=defs.BLUEZ_SERVICE,
                path=char.obj[0],
                interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                member=method,
                signature="a{sv}",
                body=[{}],
            )
        )
        assert_reply(reply)

        # we aren't actually using the write or notify, we just want the MTU
        os.close(reply.unix_fds[0])
        self._mtu_size = reply.body[1]

    async def _get_adapter_path(self) -> str:
        """Private coroutine to return the BlueZ path to the adapter this client is assigned to.

        Can be called even if no connection has been established yet.
        """
        if self._device_info:
            # If we have a BlueZ DBus object with _device_info, use what it tell us
            return self._device_info["Adapter"]
        if self._adapter:
            # If the adapter name was set in the constructor, convert to a BlueZ path
            return f"/org/bluez/{self._adapter}"

        # Fall back to the system's default Bluetooth adapter
        manager = await get_global_bluez_manager()
        return manager.get_default_adapter()

    async def _get_device_path(self) -> str:
        """Private coroutine to return the BlueZ path to the device address this client is assigned to.

        Unlike the _device_path property, this function can be called even if discovery process has not
        started and/or connection has not been established yet.
        """
        if self._device_path:
            # If we have a BlueZ DBus object, return its device path
            return self._device_path

        # Otherwise, build a new path using the adapter path and the BLE address
        adapter_path = await self._get_adapter_path()
        bluez_address = self.address.upper().replace(":", "_")
        return f"{adapter_path}/dev_{bluez_address}"

    @property
    @override
    def name(self) -> str:
        """See :meth:`bleak.BleakClient.name`."""
        if self._device_info is None:
            raise BleakError("Not connected")
        return self._device_info["Alias"]

    @property
    @override
    def mtu_size(self) -> int:
        """Get ATT MTU size for active connection"""
        if self._mtu_size is None:
            warnings.warn(
                "Using default MTU value. Call _acquire_mtu() or set _mtu_size first to avoid this warning."
            )
            return 23

        return self._mtu_size

    # GATT services methods

    async def _get_services(
        self, dangerous_use_bleak_cache: bool = False, **kwargs: Any
    ) -> BleakGATTServiceCollection:
        """Get all services registered for this GATT server.

        Args:
            dangerous_use_bleak_cache (bool): Use cached services if available.

        Returns:
           A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        if self.services is not None:
            return self.services

        manager = await get_global_bluez_manager()

        self.services = await manager.get_services(
            self._device_path, dangerous_use_bleak_cache, self._requested_services
        )

        return self.services

    # IO methods

    @override
    async def read_gatt_char(
        self, characteristic: BleakGATTCharacteristic, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to read from.

        Returns:
            (bytearray) The read data.

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        while True:
            assert self._bus

            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=characteristic.obj[0],
                    interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                    member="ReadValue",
                    signature="a{sv}",
                    body=[{}],
                )
            )

            if reply.error_name == "org.bluez.Error.InProgress":
                logger.debug("retrying characteristic ReadValue due to InProgress")
                # Avoid calling in a tight loop. There is no dbus signal to
                # indicate ready, so unfortunately, we have to poll.
                await asyncio.sleep(0.01)
                continue

            assert_reply(reply)
            break

        value = bytearray(reply.body[0])

        logger.debug(
            "Read Characteristic {0} | {1}: {2}".format(
                characteristic.uuid, characteristic.obj[0], value
            )
        )
        return value

    @override
    async def read_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.

        Returns:
            The read data.
        """
        if not self.is_connected:
            raise BleakError("Not connected")

        while True:
            assert self._bus

            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=descriptor.obj[0],
                    interface=defs.GATT_DESCRIPTOR_INTERFACE,
                    member="ReadValue",
                    signature="a{sv}",
                    body=[{}],
                )
            )

            if reply.error_name == "org.bluez.Error.InProgress":
                logger.debug("retrying descriptor ReadValue due to InProgress")
                # Avoid calling in a tight loop. There is no dbus signal to
                # indicate ready, so unfortunately, we have to poll.
                await asyncio.sleep(0.01)
                continue

            assert_reply(reply)
            break

        value = bytearray(reply.body[0])

        logger.debug(
            "Read Descriptor %s | %s: %s", descriptor.handle, descriptor.obj[0], value
        )
        return value

    @override
    async def write_gatt_char(
        self, characteristic: BleakGATTCharacteristic, data: Buffer, response: bool
    ) -> None:
        if not self.is_connected:
            raise BleakError("Not connected")

        while True:
            assert self._bus

            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=characteristic.obj[0],
                    interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                    member="WriteValue",
                    signature="aya{sv}",
                    body=[
                        bytes(data),
                        {"type": Variant("s", "request" if response else "command")},
                    ],
                )
            )

            if reply.error_name == "org.bluez.Error.InProgress":
                logger.debug("retrying characteristic WriteValue due to InProgress")
                # Avoid calling in a tight loop. There is no dbus signal to
                # indicate ready, so unfortunately, we have to poll.
                await asyncio.sleep(0.01)
                continue

            assert_reply(reply)
            break

        logger.debug(
            "Write Characteristic %s | %s: %s",
            characteristic.uuid,
            characteristic.obj[0],
            data,
        )

    @override
    async def write_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, data: Buffer
    ) -> None:
        """Perform a write operation on the specified GATT descriptor.

        Args:
            handle: The handle of the descriptor to read from.
            data: The data to send (any bytes-like object).

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        while True:
            assert self._bus

            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=descriptor.obj[0],
                    interface=defs.GATT_DESCRIPTOR_INTERFACE,
                    member="WriteValue",
                    signature="aya{sv}",
                    body=[bytes(data), {"type": Variant("s", "command")}],
                )
            )

            if reply.error_name == "org.bluez.Error.InProgress":
                logger.debug("retrying descriptor WriteValue due to InProgress")
                # Avoid calling in a tight loop. There is no dbus signal to
                # indicate ready, so unfortunately, we have to poll.
                await asyncio.sleep(0.01)
                continue

            assert_reply(reply)
            break

        logger.debug(
            "Write Descriptor %s | %s: %s", descriptor.handle, descriptor.obj[0], data
        )

    def _register_notify_fd_reader(
        self, char_path: str, fd: int, callback: NotifyCallback
    ) -> None:
        loop = asyncio.get_running_loop()

        def on_data():
            try:
                data = os.read(fd, 1024)
                if not data:
                    raise RuntimeError("Unexpected EOF on notification file handle")
            except Exception as e:
                logger.debug(
                    "AcquireNotify: Read error on fd %d: %s. Notifications have been stopped.",
                    fd,
                    e,
                )
                try:
                    loop.remove_reader(fd)
                except RuntimeError:
                    # Run loop is closed
                    pass
                try:
                    os.close(fd)
                except OSError:
                    # Bad file descriptor
                    pass
                self._notification_fds.pop(char_path, None)
                return

            callback(bytearray(data))

        loop.add_reader(fd, on_data)

    @override
    async def start_notify(
        self,
        characteristic: BleakGATTCharacteristic,
        callback: NotifyCallback,
        **kwargs: Any,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.
        """
        assert self._bus is not None

        # If using StartNotify and calling a read on the same
        # characteristic, BlueZ will return the response as
        # both a notification and read, duplicating the message.
        # Using AcquireNotify on supported characteristics avoids this.
        # However, using the preferred AcquireNotify requires that devices
        # correctly indicate "notify" and/or "indicate" properties. If they
        # don't, we fall back to StartNotify.
        use_notify_acquire = "NotifyAcquired" in characteristic.obj[1]
        logger.debug(
            'using "%s" for notifications on characteristic %d',
            "AcquireNotify" if use_notify_acquire else "StartNotify",
            characteristic.handle,
        )
        if use_notify_acquire:
            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=characteristic.obj[0],
                    interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                    member="AcquireNotify",
                    body=[{}],
                    signature="a{sv}",
                )
            )
            assert_reply(reply)

            unix_fd = reply.unix_fds[0]
            self._notification_fds[characteristic.obj[0]] = unix_fd
            self._register_notify_fd_reader(characteristic.obj[0], unix_fd, callback)
        else:
            self._notification_callbacks[characteristic.obj[0]] = callback
            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=characteristic.obj[0],
                    interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                    member="StartNotify",
                )
            )
            assert_reply(reply)

    @override
    async def stop_notify(self, characteristic: BleakGATTCharacteristic) -> None:
        """Deactivate notification/indication on a specified characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to deactivate
                notification/indication on.
        """
        if not self.is_connected:
            raise BleakError("Not connected")

        assert self._bus is not None

        if "NotifyAcquired" in characteristic.obj[1]:
            logger.debug(
                "Closing notification fd for characteristic %d", characteristic.handle
            )
            fd = self._notification_fds.pop(characteristic.obj[0], None)

            if fd is None:
                logger.debug(
                    "No notification fd found for characteristic %d",
                    characteristic.handle,
                )
            else:
                loop = asyncio.get_running_loop()
                try:
                    loop.remove_reader(fd)
                except RuntimeError:
                    # Run loop is closed
                    pass
                try:
                    os.close(fd)
                except OSError as e:
                    logger.debug("Failed to remove file descriptor %d: %s", fd, e)
        else:
            logger.debug(
                "Calling StopNotify for characteristic %d", characteristic.handle
            )
            reply = await self._bus.call(
                Message(
                    destination=defs.BLUEZ_SERVICE,
                    path=characteristic.obj[0],
                    interface=defs.GATT_CHARACTERISTIC_INTERFACE,
                    member="StopNotify",
                )
            )
            assert_reply(reply)
            self._notification_callbacks.pop(characteristic.obj[0], None)



import sys
from typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

import logging
from typing import Any, Literal, Optional
from warnings import warn

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override

import objc
from CoreBluetooth import CBPeripheral
from Foundation import NSBundle, NSDictionary

from bleak.args.corebluetooth import CBScannerArgs as _CBScannerArgs
from bleak.backends.corebluetooth.CentralManagerDelegate import CentralManagerDelegate
from bleak.backends.corebluetooth.utils import cb_uuid_to_str
from bleak.backends.scanner import (
    AdvertisementData,
    AdvertisementDataCallback,
    BaseBleakScanner,
)
from bleak.exc import BleakError

logger = logging.getLogger(__name__)


def __getattr__(name: str):
    if name == "CBScannerArgs":
        warn(
            "importing CBScannerArgs from bleak.backends.corebluetooth.scanner is deprecated, use bleak.args.corebluetooth instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return _CBScannerArgs
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class BleakScannerCoreBluetooth(BaseBleakScanner):
    """The native macOS Bleak BLE Scanner.

    Documentation:
    https://developer.apple.com/documentation/corebluetooth/cbcentralmanager

    CoreBluetooth doesn't explicitly use Bluetooth addresses to identify peripheral
    devices because private devices may obscure their Bluetooth addresses. To cope
    with this, CoreBluetooth utilizes UUIDs for each peripheral. Bleak uses
    this for the BLEDevice address on macOS.

    Args:
        detection_callback:
            Optional function that will be called each time a device is
            discovered or advertising data has changed.
        service_uuids:
            Optional list of service UUIDs to filter on. Only advertisements
            containing this advertising data will be received. Required on
            macOS >= 12.0, < 12.3 (unless you create an app with ``py2app``).
        scanning_mode:
            Set to ``"passive"`` to avoid the ``"active"`` scanning mode. Not
            supported on macOS! Will raise :class:`BleakError` if set to
            ``"passive"``
        **timeout (float):
             The scanning timeout to be used, in case of missing
            ``stopScan_`` method.
    """

    def __init__(
        self,
        detection_callback: Optional[AdvertisementDataCallback],
        service_uuids: Optional[list[str]],
        scanning_mode: Literal["active", "passive"],
        *,
        cb: _CBScannerArgs,
        **kwargs: Any,
    ):
        super(BleakScannerCoreBluetooth, self).__init__(
            detection_callback, service_uuids
        )

        self._use_bdaddr = cb.get("use_bdaddr", False)

        if scanning_mode == "passive":
            raise BleakError("macOS does not support passive scanning")

        self._manager = CentralManagerDelegate()
        self._timeout: float = kwargs.get("timeout", 5.0)
        if (
            objc.macos_available(12, 0)
            and not objc.macos_available(12, 3)
            and not self._service_uuids
        ):
            # See https://github.com/hbldh/bleak/issues/720
            if NSBundle.mainBundle().bundleIdentifier() == "org.python.python":
                logger.error(
                    "macOS 12.0, 12.1 and 12.2 require non-empty service_uuids kwarg, otherwise no advertisement data will be received"
                )

    @override
    async def start(self) -> None:
        await self._manager.wait_until_ready()

        self.seen_devices = {}

        def callback(p: CBPeripheral, a: NSDictionary, r: int) -> None:

            service_uuids = [
                cb_uuid_to_str(u) for u in a.get("kCBAdvDataServiceUUIDs", [])
            ]

            if not self.is_allowed_uuid(service_uuids):
                return

            # Process service data
            service_data_dict_raw = a.get("kCBAdvDataServiceData", {})
            service_data = {
                cb_uuid_to_str(k): bytes(v) for k, v in service_data_dict_raw.items()
            }

            # Process manufacturer data into a more friendly format
            manufacturer_binary_data = a.get("kCBAdvDataManufacturerData")
            manufacturer_data: dict[int, bytes] = {}
            if manufacturer_binary_data:
                manufacturer_id = int.from_bytes(
                    manufacturer_binary_data[0:2], byteorder="little"
                )
                manufacturer_value = bytes(manufacturer_binary_data[2:])
                manufacturer_data[manufacturer_id] = manufacturer_value

            # set tx_power data if available
            tx_power = a.get("kCBAdvDataTxPowerLevel")

            advertisement_data = AdvertisementData(
                local_name=a.get("kCBAdvDataLocalName"),
                manufacturer_data=manufacturer_data,
                service_data=service_data,
                service_uuids=service_uuids,
                tx_power=tx_power,
                rssi=r,
                platform_data=(p, a, r),
            )

            if self._use_bdaddr:
                # HACK: retrieveAddressForPeripheral_ is undocumented but seems to do the trick
                address_bytes = cast(
                    Optional[bytes],
                    self._manager.central_manager.retrieveAddressForPeripheral_(p),  # type: ignore
                )
                if address_bytes is None:
                    logger.debug(
                        "Could not get Bluetooth address for %s. Ignoring this device.",
                        p.identifier().UUIDString(),
                    )
                    return
                address = address_bytes.hex(":").upper()
            else:
                address = p.identifier().UUIDString()

            device = self.create_or_update_device(
                p.identifier().UUIDString(),
                address,
                p.name(),
                (p, self._manager),
                advertisement_data,
            )

            self.call_detection_callbacks(device, advertisement_data)

        self._manager.callbacks[id(self)] = callback
        await self._manager.start_scan(self._service_uuids)

    @override
    async def stop(self) -> None:
        await self._manager.stop_scan()
        self._manager.callbacks.pop(id(self), None)



"""
-----------------------
BlueZ backend arguments
-----------------------
"""

from typing import NamedTuple, TypedDict, Union

from bleak.assigned_numbers import AdvertisementDataType


class BlueZDiscoveryFilters(TypedDict, total=False):
    """
    Dictionary of arguments for the ``org.bluez.Adapter1.SetDiscoveryFilter``
    D-Bus method.

    https://github.com/bluez/bluez/blob/master/doc/org.bluez.Adapter.rst#void-setdiscoveryfilterdict-filter
    """

    UUIDs: list[str]
    """
    Filter by service UUIDs, empty means match _any_ UUID.

    Normally, the ``service_uuids`` argument of :class:`bleak.BleakScanner`
    is used instead.
    """
    RSSI: int
    """
    RSSI threshold value.
    """
    Pathloss: int
    """
    Pathloss threshold value.
    """
    Transport: str
    """
    Transport parameter determines the type of scan.

    This should not be used since it is required to be set to ``"le"``.
    """
    DuplicateData: bool
    """
    Disables duplicate detection of advertisement data.

    This does not affect the ``Filter Duplicates`` parameter of the ``LE Set Scan Enable``
    HCI command to the Bluetooth adapter!

    Although the default value for BlueZ is ``True``, Bleak sets this to ``False`` by default.
    """
    Discoverable: bool
    """
    Make adapter discoverable while discovering,
    if the adapter is already discoverable setting
    this filter won't do anything.
    """
    Pattern: str
    """
    Discover devices where the pattern matches
    either the prefix of the address or
    device name which is convenient way to limited
    the number of device objects created during a
    discovery.
    """


class OrPattern(NamedTuple):
    """
    BlueZ advertisement monitor or-pattern.

    https://github.com/bluez/bluez/blob/master/doc/org.bluez.AdvertisementMonitor.rst#arrayuint8-uint8-arraybyte-patterns-read-only-optional
    """

    start_position: int
    ad_data_type: AdvertisementDataType
    content_of_pattern: bytes


# Windows has a similar structure, so we allow generic tuple for cross-platform compatibility
OrPatternLike = Union[OrPattern, tuple[int, AdvertisementDataType, bytes]]


class BlueZScannerArgs(TypedDict, total=False):
    """
    :class:`BleakScanner` args that are specific to the BlueZ backend.
    """

    filters: BlueZDiscoveryFilters
    """
    Filters to pass to the adapter SetDiscoveryFilter D-Bus method.

    Only used for active scanning.
    """

    or_patterns: list[OrPatternLike]
    """
    Or patterns to pass to the AdvertisementMonitor1 D-Bus interface.

    Only used for passive scanning.
    """



"""
-----------------------
WinRT backend arguments
-----------------------
"""

from typing import Literal, TypedDict


class WinRTClientArgs(TypedDict, total=False):
    """
    Windows-specific arguments for :class:`BleakClient`.
    """

    address_type: Literal["public", "random"]
    """
    Can either be ``"public"`` or ``"random"``, depending on the required address
    type needed to connect to your device.
    """

    use_cached_services: bool
    """
    ``True`` allows Windows to fetch the services, characteristics and descriptors
    from the Windows cache instead of reading them from the device. Can be very
    much faster for known, unchanging devices, but not recommended for DIY peripherals
    where the GATT layout can change between connections.

    ``False`` will force the attribute database to be read from the remote device
    instead of using the OS cache.

    If omitted, the OS Bluetooth stack will do what it thinks is best.
    """



"""

PeripheralDelegate

Created by kevincar <kevincarrolldavis@gmail.com>

"""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "darwin":
        assert False, "This backend is only available on macOS"

import asyncio
import itertools
import logging
from collections.abc import Iterable
from typing import Any, Optional

if sys.version_info < (3, 11):
    from async_timeout import timeout as async_timeout
    from typing_extensions import Self
else:
    from asyncio import timeout as async_timeout
    from typing import Self

import objc
from CoreBluetooth import (
    CBUUID,
    CBCharacteristic,
    CBCharacteristicWriteType,
    CBCharacteristicWriteWithResponse,
    CBDescriptor,
    CBPeripheral,
    CBService,
)
from Foundation import NSUUID, NSArray, NSData, NSError, NSObject

from bleak.args.corebluetooth import NotificationDiscriminator
from bleak.backends.client import NotifyCallback
from bleak.exc import BleakError

# logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)


CBPeripheralDelegate = objc.protocolNamed("CBPeripheralDelegate")


class ObjcPeripheralDelegate(NSObject, protocols=[CBPeripheralDelegate]):
    """
    CoreBluetooth peripheral manager delegate for bridging callbacks to asyncio.
    """

    def initWithPyDelegate_(self, py_delegate: PeripheralDelegate) -> Optional[Self]:
        """macOS init function for NSObject"""
        self = objc.super(ObjcPeripheralDelegate, self).init()

        if self is None:
            return None

        self.py_delegate = py_delegate

        return self

    # Protocol Functions

    def peripheral_didDiscoverServices_(
        self, peripheral: CBPeripheral, error: Optional[NSError]
    ) -> None:
        logger.debug("peripheral_didDiscoverServices_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_discover_services,
                peripheral,
                peripheral.services(),
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didDiscoverIncludedServicesForService_error_(
        self, peripheral: CBPeripheral, service: CBService, error: Optional[NSError]
    ) -> None:
        logger.debug("peripheral_didDiscoverIncludedServicesForService_error_")
        # Currently not used in Bleak

    def peripheral_didDiscoverCharacteristicsForService_error_(
        self, peripheral: CBPeripheral, service: CBService, error: Optional[NSError]
    ) -> None:
        logger.debug("peripheral_didDiscoverCharacteristicsForService_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_discover_characteristics_for_service,
                peripheral,
                service,
                service.characteristics(),
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didDiscoverDescriptorsForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didDiscoverDescriptorsForCharacteristic_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_discover_descriptors_for_characteristic,
                peripheral,
                characteristic,
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didUpdateValueForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didUpdateValueForCharacteristic_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_update_value_for_characteristic,
                peripheral,
                characteristic,
                characteristic.value(),
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didUpdateValueForDescriptor_error_(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didUpdateValueForDescriptor_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_update_value_for_descriptor,
                peripheral,
                descriptor,
                descriptor.value(),
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didWriteValueForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didWriteValueForCharacteristic_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_write_value_for_characteristic,
                peripheral,
                characteristic,
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didWriteValueForDescriptor_error_(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didWriteValueForDescriptor_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_write_value_for_descriptor,
                peripheral,
                descriptor,
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheralIsReadyToSendWriteWithoutResponse_(
        self, peripheral: CBPeripheral
    ) -> None:
        logger.debug("peripheralIsReadyToSendWriteWithoutResponse_")
        # Currently not used in Bleak

    def peripheral_didUpdateNotificationStateForCharacteristic_error_(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didUpdateNotificationStateForCharacteristic_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_update_notification_for_characteristic,
                peripheral,
                characteristic,
                error,
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didReadRSSI_error_(
        self,
        peripheral: CBPeripheral,
        rssi: int,
        error: Optional[NSError],
    ) -> None:
        logger.debug("peripheral_didReadRSSI_error_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_read_rssi, peripheral, rssi, error
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    # Bleak currently doesn't use the callbacks below other than for debug logging

    def peripheralDidUpdateName_(self, peripheral: CBPeripheral) -> None:
        logger.debug("peripheralDidUpdateName_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_update_name, peripheral, peripheral.name()
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)

    def peripheral_didModifyServices_(
        self, peripheral: CBPeripheral, invalidatedServices: NSArray[CBService]
    ) -> None:
        logger.debug("peripheral_didModifyServices_")

        try:
            self.py_delegate.event_loop.call_soon_threadsafe(
                self.py_delegate.did_modify_services, peripheral, invalidatedServices
            )
        except RuntimeError as e:
            # Likely caused by loop being closed
            logger.debug("unraisable exception", exc_info=e)


class PeripheralDelegate:
    """macOS conforming python class for managing the PeripheralDelegate for BLE"""

    def __init__(self, peripheral: CBPeripheral) -> None:
        delegate = ObjcPeripheralDelegate.alloc().initWithPyDelegate_(self)
        assert delegate is not None
        self.objc_delegate = delegate

        self.peripheral = peripheral
        self.peripheral.setDelegate_(self.objc_delegate)

        self.event_loop = asyncio.get_running_loop()
        self._services_discovered_future = self.event_loop.create_future()

        self._service_characteristic_discovered_futures: dict[
            int, asyncio.Future[NSArray[CBCharacteristic]]
        ] = {}
        self._characteristic_descriptor_discover_futures: dict[
            int, asyncio.Future[None]
        ] = {}

        self._characteristic_read_futures: dict[int, asyncio.Future[NSData]] = {}
        self._characteristic_write_futures: dict[int, asyncio.Future[None]] = {}

        self._descriptor_read_futures: dict[int, asyncio.Future[NSObject]] = {}
        self._descriptor_write_futures: dict[int, asyncio.Future[None]] = {}

        self._characteristic_notify_change_futures: dict[int, asyncio.Future[None]] = {}
        self._characteristic_notify_callbacks: dict[int, NotifyCallback] = {}
        self._characteristic_notification_discriminators: dict[
            int, Optional[NotificationDiscriminator]
        ] = {}

        self._read_rssi_futures: dict[NSUUID, asyncio.Future[int]] = {}

    def futures(self) -> Iterable[asyncio.Future[Any]]:
        """
        Gets all futures for this delegate.

        These can be used to handle any pending futures when a peripheral is disconnected.
        """
        services_discovered_future = (
            (self._services_discovered_future,)
            if hasattr(self, "_services_discovered_future")
            else ()
        )

        return itertools.chain(
            services_discovered_future,
            self._service_characteristic_discovered_futures.values(),
            self._characteristic_descriptor_discover_futures.values(),
            self._characteristic_read_futures.values(),
            self._characteristic_write_futures.values(),
            self._descriptor_read_futures.values(),
            self._descriptor_write_futures.values(),
            self._characteristic_notify_change_futures.values(),
            self._read_rssi_futures.values(),
        )

    async def discover_services(
        self, services: Optional[NSArray[CBUUID]] = None
    ) -> NSArray[CBService]:
        future = self.event_loop.create_future()

        self._services_discovered_future = future
        try:
            self.peripheral.discoverServices_(services)
            return await future
        finally:
            del self._services_discovered_future

    async def discover_characteristics(
        self, service: CBService
    ) -> NSArray[CBCharacteristic]:
        future = self.event_loop.create_future()

        self._service_characteristic_discovered_futures[service.startHandle()] = future
        try:
            self.peripheral.discoverCharacteristics_forService_(None, service)
            return await future
        finally:
            del self._service_characteristic_discovered_futures[service.startHandle()]

    async def discover_descriptors(
        self, characteristic: CBCharacteristic
    ) -> NSArray[CBDescriptor]:
        future = self.event_loop.create_future()

        self._characteristic_descriptor_discover_futures[characteristic.handle()] = (
            future
        )
        try:
            self.peripheral.discoverDescriptorsForCharacteristic_(characteristic)
            await future
        finally:
            del self._characteristic_descriptor_discover_futures[
                characteristic.handle()
            ]

        return characteristic.descriptors()

    async def read_characteristic(
        self,
        characteristic: CBCharacteristic,
        use_cached: bool,
        timeout: int = 20,
    ) -> NSData:
        value = characteristic.value()
        if value is not None and use_cached:
            return value

        future = self.event_loop.create_future()

        self._characteristic_read_futures[characteristic.handle()] = future

        try:
            self.peripheral.readValueForCharacteristic_(characteristic)
            async with async_timeout(timeout):
                return await future
        finally:
            del self._characteristic_read_futures[characteristic.handle()]

    async def read_descriptor(
        self, descriptor: CBDescriptor, use_cached: bool = True
    ) -> Any:
        value = descriptor.value()
        if value is not None and use_cached:
            return value

        future = self.event_loop.create_future()

        self._descriptor_read_futures[descriptor.handle()] = future
        try:
            self.peripheral.readValueForDescriptor_(descriptor)
            return await future
        finally:
            del self._descriptor_read_futures[descriptor.handle()]

    async def write_characteristic(
        self,
        characteristic: CBCharacteristic,
        value: NSData,
        response: CBCharacteristicWriteType,
    ) -> None:
        # in CoreBluetooth there is no indication of success or failure of
        # CBCharacteristicWriteWithoutResponse
        if response == CBCharacteristicWriteWithResponse:
            future = self.event_loop.create_future()

            self._characteristic_write_futures[characteristic.handle()] = future
            try:
                self.peripheral.writeValue_forCharacteristic_type_(
                    value, characteristic, response
                )
                await future
            finally:
                del self._characteristic_write_futures[characteristic.handle()]
        else:
            self.peripheral.writeValue_forCharacteristic_type_(
                value, characteristic, response
            )

    async def write_descriptor(self, descriptor: CBDescriptor, value: NSData) -> None:
        future = self.event_loop.create_future()

        self._descriptor_write_futures[descriptor.handle()] = future
        try:
            self.peripheral.writeValue_forDescriptor_(value, descriptor)
            await future
        finally:
            del self._descriptor_write_futures[descriptor.handle()]

    async def start_notifications(
        self,
        characteristic: CBCharacteristic,
        callback: NotifyCallback,
        notification_discriminator: Optional[NotificationDiscriminator] = None,
    ) -> None:
        c_handle = characteristic.handle()
        if c_handle in self._characteristic_notify_callbacks:
            raise ValueError("Characteristic notifications already started")

        self._characteristic_notify_callbacks[c_handle] = callback
        self._characteristic_notification_discriminators[c_handle] = (
            notification_discriminator
        )

        future = self.event_loop.create_future()

        self._characteristic_notify_change_futures[c_handle] = future
        try:
            self.peripheral.setNotifyValue_forCharacteristic_(True, characteristic)
            await future
        finally:
            del self._characteristic_notify_change_futures[c_handle]

    async def stop_notifications(self, characteristic: CBCharacteristic) -> None:
        c_handle = characteristic.handle()
        if c_handle not in self._characteristic_notify_callbacks:
            raise ValueError("Characteristic notification never started")

        future = self.event_loop.create_future()

        self._characteristic_notify_change_futures[c_handle] = future
        try:
            self.peripheral.setNotifyValue_forCharacteristic_(False, characteristic)
            await future
        finally:
            del self._characteristic_notify_change_futures[c_handle]

        self._characteristic_notify_callbacks.pop(c_handle)
        self._characteristic_notification_discriminators.pop(c_handle)

    async def read_rssi(self) -> int:
        future = self.event_loop.create_future()

        self._read_rssi_futures[self.peripheral.identifier()] = future
        try:
            self.peripheral.readRSSI()
            return await future
        finally:
            del self._read_rssi_futures[self.peripheral.identifier()]

    # Protocol Functions

    def did_discover_services(
        self,
        peripheral: CBPeripheral,
        services: NSArray[CBService],
        error: Optional[NSError],
    ) -> None:
        future = self._services_discovered_future
        if error is not None:
            exception = BleakError(f"Failed to discover services {error}")
            future.set_exception(exception)
        else:
            logger.debug("Services discovered")
            future.set_result(services)

    def did_discover_characteristics_for_service(
        self,
        peripheral: CBPeripheral,
        service: CBService,
        characteristics: NSArray[CBCharacteristic],
        error: Optional[NSError],
    ) -> None:
        future = self._service_characteristic_discovered_futures.get(
            service.startHandle()
        )
        if not future:
            logger.debug(
                f"Unexpected event didDiscoverCharacteristicsForService for {service.startHandle()}"
            )
            return
        if error is not None:
            exception = BleakError(
                f"Failed to discover characteristics for service {service.startHandle()}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug("Characteristics discovered")
            future.set_result(characteristics)

    def did_discover_descriptors_for_characteristic(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        future = self._characteristic_descriptor_discover_futures.get(
            characteristic.handle()
        )
        if not future:
            logger.warning(
                f"Unexpected event didDiscoverDescriptorsForCharacteristic for {characteristic.handle()}"
            )
            return
        if error is not None:
            exception = BleakError(
                f"Failed to discover descriptors for characteristic {characteristic.handle()}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug(f"Descriptor discovered {characteristic.handle()}")
            future.set_result(None)

    def did_update_value_for_characteristic(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        value: Optional[NSData],
        error: Optional[NSError],
    ) -> None:
        c_handle = characteristic.handle()

        future = self._characteristic_read_futures.get(c_handle)

        # If error is set, then we know this was a read response.
        # Otherwise, if there is a pending read request, we can't tell if this is a read response or notification.
        # If the user provided a notification discriminator, we can use that to
        # identify if this callback is due to a notification by analyzing the value.
        # If not, and there is a future (pending read request), we assume it is a read response but can't know for sure.
        if not error:
            assert value is not None

            notification_discriminator = (
                self._characteristic_notification_discriminators.get(c_handle)
            )
            if not future or (
                notification_discriminator and notification_discriminator(bytes(value))
            ):
                notify_callback = self._characteristic_notify_callbacks.get(c_handle)

                if notify_callback:
                    notify_callback(bytearray(value))
                    return

        if not future:
            logger.warning(
                "Unexpected event didUpdateValueForCharacteristic for 0x%04x with value: %r and error: %r",
                c_handle,
                value,
                error,
            )
            return

        if error is not None:
            exception = BleakError(f"Failed to read characteristic {c_handle}: {error}")
            future.set_exception(exception)
        else:
            logger.debug("Read characteristic value")
            assert value is not None
            future.set_result(value)

    def did_update_value_for_descriptor(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        value: Optional[Any],
        error: Optional[NSError],
    ) -> None:
        future = self._descriptor_read_futures.get(descriptor.handle())
        if not future:
            logger.warning("Unexpected event didUpdateValueForDescriptor")
            return
        if error is not None:
            exception = BleakError(
                f"Failed to read descriptor {descriptor.handle()}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug("Read descriptor value")
            assert value is not None
            future.set_result(value)

    def did_write_value_for_characteristic(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        future = self._characteristic_write_futures.get(characteristic.handle(), None)
        if not future:
            return  # event only expected on write with response
        if error is not None:
            exception = BleakError(
                f"Failed to write characteristic {characteristic.handle()}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug("Write Characteristic Value")
            future.set_result(None)

    def did_write_value_for_descriptor(
        self,
        peripheral: CBPeripheral,
        descriptor: CBDescriptor,
        error: Optional[NSError],
    ) -> None:
        future = self._descriptor_write_futures.get(descriptor.handle())
        if not future:
            logger.warning("Unexpected event didWriteValueForDescriptor")
            return
        if error is not None:
            exception = BleakError(
                f"Failed to write descriptor {descriptor.handle()}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug("Write Descriptor Value")
            future.set_result(None)

    def did_update_notification_for_characteristic(
        self,
        peripheral: CBPeripheral,
        characteristic: CBCharacteristic,
        error: Optional[NSError],
    ) -> None:
        c_handle = characteristic.handle()
        future = self._characteristic_notify_change_futures.get(c_handle)
        if not future:
            logger.warning(
                "Unexpected event didUpdateNotificationStateForCharacteristic"
            )
            return
        if error is not None:
            exception = BleakError(
                f"Failed to update the notification status for characteristic {c_handle}: {error}"
            )
            future.set_exception(exception)
        else:
            logger.debug("Character Notify Update")
            future.set_result(None)

    def did_read_rssi(
        self, peripheral: CBPeripheral, rssi: int, error: Optional[NSError]
    ) -> None:
        future = self._read_rssi_futures.get(peripheral.identifier(), None)

        if not future:
            logger.warning("Unexpected event did_read_rssi")
            return

        if error is not None:
            exception = BleakError(f"Failed to read RSSI: {error}")
            future.set_exception(exception)
        else:
            future.set_result(rssi)

    # Bleak currently doesn't use the callbacks below other than for debug logging

    def did_update_name(self, peripheral: CBPeripheral, name: str) -> None:
        logger.debug(f"name of {peripheral.identifier()} changed to {name}")

    def did_modify_services(
        self, peripheral: CBPeripheral, invalidated_services: NSArray[CBService]
    ) -> None:
        logger.debug(
            f"{peripheral.identifier()} invalidated services: {invalidated_services}"
        )



import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    if sys.platform != "win32":
        assert False, "This backend is only available on Windows"

import asyncio
import logging
from typing import Literal, NamedTuple, Optional
from uuid import UUID

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override

from winrt.windows.devices.bluetooth import BluetoothAdapter
from winrt.windows.devices.bluetooth.advertisement import (
    BluetoothLEAdvertisementReceivedEventArgs,
    BluetoothLEAdvertisementType,
    BluetoothLEAdvertisementWatcher,
    BluetoothLEAdvertisementWatcherStatus,
    BluetoothLEAdvertisementWatcherStoppedEventArgs,
    BluetoothLEScanningMode,
)
from winrt.windows.devices.radios import RadioState
from winrt.windows.foundation import EventRegistrationToken

from bleak.assigned_numbers import AdvertisementDataType
from bleak.backends.scanner import (
    AdvertisementData,
    AdvertisementDataCallback,
    BaseBleakScanner,
)
from bleak.backends.winrt.util import assert_mta
from bleak.exc import (
    BleakBluetoothNotAvailableError,
    BleakBluetoothNotAvailableReason,
    BleakError,
)
from bleak.uuids import normalize_uuid_str

logger = logging.getLogger(__name__)


def _format_bdaddr(a: int) -> str:
    return ":".join(f"{x:02X}" for x in a.to_bytes(6, byteorder="big"))


def _format_event_args(e: BluetoothLEAdvertisementReceivedEventArgs) -> str:
    try:
        return f"{_format_bdaddr(e.bluetooth_address)}: {e.advertisement.local_name}"
    except Exception:
        return _format_bdaddr(e.bluetooth_address)


class RawAdvData(NamedTuple):
    """
    Platform-specific advertisement data.

    Windows does not combine advertising data with type SCAN_RSP with other
    advertising data like other platforms, so we have to do it ourselves.
    """

    adv: Optional[BluetoothLEAdvertisementReceivedEventArgs]
    """
    The advertisement data received from the BluetoothLEAdvertisementWatcher.Received event.
    """
    scan: Optional[BluetoothLEAdvertisementReceivedEventArgs]
    """
    The scan response for the same device as *adv*.
    """


class BleakScannerWinRT(BaseBleakScanner):
    """The native Windows Bleak BLE Scanner.

    Implemented using `Python/WinRT <https://github.com/Microsoft/xlang/tree/master/src/package/pywinrt/projection/>`_.

    Args:
        detection_callback:
            Optional function that will be called each time a device is
            discovered or advertising data has changed.
        service_uuids:
            Optional list of service UUIDs to filter on. Only advertisements
            containing this advertising data will be received.
        scanning_mode:
            Set to ``"passive"`` to avoid the ``"active"`` scanning mode.

    """

    def __init__(
        self,
        detection_callback: Optional[AdvertisementDataCallback],
        service_uuids: Optional[list[str]],
        scanning_mode: Literal["active", "passive"],
        **kwargs: Any,
    ):
        super(BleakScannerWinRT, self).__init__(detection_callback, service_uuids)

        self.watcher: Optional[BluetoothLEAdvertisementWatcher] = None
        self._advertisement_pairs: dict[str, RawAdvData] = {}
        self._stopped_event: Optional[asyncio.Event] = None

        # case insensitivity is for backwards compatibility on Windows only
        if scanning_mode.lower() == "passive":
            self._scanning_mode = BluetoothLEScanningMode.PASSIVE
        else:
            self._scanning_mode = BluetoothLEScanningMode.ACTIVE

        # Unfortunately, due to the way Windows handles filtering, we can't
        # make use of the service_uuids filter here. If we did we would only
        # get the advertisement data or the scan data, but not both, so would
        # miss out on other essential data. Advanced users can pass their own
        # filters though if they want to.
        self._signal_strength_filter = kwargs.get("SignalStrengthFilter", None)
        self._advertisement_filter = kwargs.get("AdvertisementFilter", None)

        self._received_token: Optional[EventRegistrationToken] = None
        self._stopped_token: Optional[EventRegistrationToken] = None

    def _received_handler(
        self,
        sender: BluetoothLEAdvertisementWatcher,
        event_args: BluetoothLEAdvertisementReceivedEventArgs,
    ):
        """Callback for AdvertisementWatcher.Received"""
        # TODO: Cannot check for if sender == self.watcher in winrt?
        logger.debug("Received %s.", _format_event_args(event_args))

        # REVISIT: if scanning filters with BluetoothSignalStrengthFilter.OutOfRangeTimeout
        # are in place, an RSSI of -127 means that the device has gone out of range and should
        # be removed from the list of seen devices instead of processing the advertisement data.
        # https://learn.microsoft.com/en-us/uwp/api/windows.devices.bluetooth.bluetoothsignalstrengthfilter.outofrangetimeout

        bdaddr = _format_bdaddr(event_args.bluetooth_address)

        # Unlike other platforms, Windows does not combine advertising data for
        # us (regular advertisement + scan response) so we have to do it manually.

        # get the previous advertising data/scan response pair or start a new one
        raw_data = self._advertisement_pairs.get(bdaddr, RawAdvData(None, None))

        # update the advertising data depending on the advertising data type
        if event_args.advertisement_type == BluetoothLEAdvertisementType.SCAN_RESPONSE:
            raw_data = RawAdvData(raw_data.adv, event_args)
        else:
            raw_data = RawAdvData(event_args, raw_data.scan)

        self._advertisement_pairs[bdaddr] = raw_data

        uuids: list[str] = []
        mfg_data = {}
        service_data = {}
        local_name = None
        tx_power = None

        for args in filter(lambda d: d is not None, raw_data):
            assert args

            for u in args.advertisement.service_uuids:
                uuids.append(str(u))

            for m in args.advertisement.manufacturer_data:
                mfg_data[m.company_id] = bytes(m.data)

            # local name is empty string rather than None if not present
            if args.advertisement.local_name:
                local_name = args.advertisement.local_name

            try:
                if args.transmit_power_level_in_dbm is not None:
                    tx_power = args.transmit_power_level_in_dbm
            except AttributeError:
                # the transmit_power_level_in_d_bm property was introduce in
                # Windows build 19041 so we have a fallback for older versions
                for section in args.advertisement.get_sections_by_type(
                    AdvertisementDataType.TX_POWER_LEVEL
                ):
                    tx_power = bytes(section.data)[0]

            # Decode service data
            for section in args.advertisement.get_sections_by_type(
                AdvertisementDataType.SERVICE_DATA_UUID16
            ):
                data = bytes(section.data)
                service_data[normalize_uuid_str(f"{data[1]:02x}{data[0]:02x}")] = data[
                    2:
                ]
            for section in args.advertisement.get_sections_by_type(
                AdvertisementDataType.SERVICE_DATA_UUID32
            ):
                data = bytes(section.data)
                service_data[
                    normalize_uuid_str(
                        f"{data[3]:02x}{data[2]:02x}{data[1]:02x}{data[0]:02x}"
                    )
                ] = data[4:]
            for section in args.advertisement.get_sections_by_type(
                AdvertisementDataType.SERVICE_DATA_UUID128
            ):
                data = bytes(section.data)
                service_data[str(UUID(bytes=bytes(data[15::-1])))] = data[16:]

        if not self.is_allowed_uuid(uuids):
            return

        # Use the BLEDevice to populate all the fields for the advertisement data to return
        advertisement_data = AdvertisementData(
            local_name=local_name,
            manufacturer_data=mfg_data,
            service_data=service_data,
            service_uuids=uuids,
            tx_power=tx_power,
            rssi=event_args.raw_signal_strength_in_dbm,
            platform_data=(sender, raw_data),
        )

        device = self.create_or_update_device(
            bdaddr, bdaddr, local_name, raw_data, advertisement_data
        )

        self.call_detection_callbacks(device, advertisement_data)

    def _stopped_handler(
        self,
        sender: BluetoothLEAdvertisementWatcher,
        e: BluetoothLEAdvertisementWatcherStoppedEventArgs,
    ) -> None:
        logger.debug(
            "%s devices found. Watcher status: %r.",
            len(self.seen_devices),
            sender.status,
        )
        assert self._stopped_event
        self._stopped_event.set()

    @override
    async def start(self) -> None:
        if self.watcher:
            raise BleakError("Scanner already started")

        # Callbacks for WinRT async methods will never happen in STA mode if
        # there is nothing pumping a Windows message loop.
        await assert_mta()

        adapter = await BluetoothAdapter.get_default_async()
        if adapter is None:
            raise BleakBluetoothNotAvailableError(
                "No Bluetooth adapter found",
                BleakBluetoothNotAvailableReason.NO_BLUETOOTH,
            )

        if not adapter.is_central_role_supported:
            raise BleakBluetoothNotAvailableError(
                "BLE 'central' role not supported on this adapter",
                BleakBluetoothNotAvailableReason.NO_BLE_CENTRAL_ROLE,
            )

        radio = await adapter.get_radio_async()
        if radio.state != RadioState.ON:
            raise BleakBluetoothNotAvailableError(
                "Bluetooth radio is not powered on. Turn on Bluetooth and try again.",
                BleakBluetoothNotAvailableReason.POWERED_OFF,
            )

        # start with fresh list of discovered devices
        self.seen_devices = {}
        self._advertisement_pairs.clear()

        self.watcher = BluetoothLEAdvertisementWatcher()
        self.watcher.scanning_mode = self._scanning_mode
        # BlueZ and CoreBluetooth don't allow controlling this and always enabled it, so do the same here
        self.watcher.allow_extended_advertisements = True

        event_loop = asyncio.get_running_loop()
        self._stopped_event = asyncio.Event()

        def on_received(
            sender: BluetoothLEAdvertisementWatcher,
            args: BluetoothLEAdvertisementReceivedEventArgs,
        ) -> None:
            event_loop.call_soon_threadsafe(self._received_handler, sender, args)

        self._received_token = self.watcher.add_received(on_received)

        def on_stopped(
            sender: BluetoothLEAdvertisementWatcher,
            args: BluetoothLEAdvertisementWatcherStoppedEventArgs,
        ) -> None:
            event_loop.call_soon_threadsafe(self._stopped_handler, sender, args)

        self._stopped_token = self.watcher.add_stopped(on_stopped)

        if self._signal_strength_filter is not None:
            self.watcher.signal_strength_filter = self._signal_strength_filter
        if self._advertisement_filter is not None:
            self.watcher.advertisement_filter = self._advertisement_filter

        self.watcher.start()

        # no events for status changes, so we have to poll :-(
        while self.watcher.status == BluetoothLEAdvertisementWatcherStatus.CREATED:
            await asyncio.sleep(0.01)

        if self.watcher.status == BluetoothLEAdvertisementWatcherStatus.ABORTED:
            raise BleakError("Failed to start scanner. Is Bluetooth turned on?")

        if self.watcher.status != BluetoothLEAdvertisementWatcherStatus.STARTED:
            raise BleakError(f"Unexpected watcher status: {self.watcher.status.name}")

    @override
    async def stop(self) -> None:
        assert self.watcher
        assert self._stopped_event
        assert self._received_token
        assert self._stopped_token

        self.watcher.stop()

        if self.watcher.status == BluetoothLEAdvertisementWatcherStatus.STOPPING:
            await self._stopped_event.wait()
        else:
            logger.debug(
                "skipping waiting for stop because status is %r",
                self.watcher.status,
            )

        try:
            self.watcher.remove_received(self._received_token)
            self.watcher.remove_stopped(self._stopped_token)
        except Exception as e:
            logger.debug("Could not remove event handlers: %s", e)

        self._stopped_token = None
        self._received_token = None

        self.watcher = None



# -*- coding: utf-8 -*-
# Created on 2019-03-19 by hbldh <henrik.blidh@nedomkull.com>
"""
Gatt Service Collection class and interface class for the Bleak representation of a GATT Service.
"""
import logging
from collections.abc import Iterator
from typing import Any, Optional, Union, cast
from uuid import UUID

from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.exc import BleakError
from bleak.uuids import normalize_uuid_str, uuidstr_to_str

logger = logging.getLogger(__name__)


class BleakGATTService:
    """The Bleak representation of a GATT Service."""

    def __init__(self, obj: Any, handle: int, uuid: str) -> None:
        self.obj = obj
        self._handle = handle
        self._uuid = uuid
        self._characteristics: dict[int, BleakGATTCharacteristic] = {}

    def __str__(self) -> str:
        return f"{self.uuid} (Handle: {self.handle}): {self.description}"

    @property
    def handle(self) -> int:
        """The handle of this service"""
        return self._handle

    @property
    def uuid(self) -> str:
        """The UUID to this service"""
        return self._uuid

    @property
    def description(self) -> str:
        """String description for this service"""
        return uuidstr_to_str(self.uuid)

    @property
    def characteristics(self) -> list[BleakGATTCharacteristic]:
        """List of characteristics for this service"""
        return list(self._characteristics.values())

    def add_characteristic(self, characteristic: BleakGATTCharacteristic) -> None:
        """Add a :py:class:`~BleakGATTCharacteristic` to the service.

        Should not be used by end user, but rather by `bleak` itself.
        """
        if characteristic.handle in self._characteristics:
            raise BleakError(
                "The characteristic '%s' is already present in this BleakGATTService!",
                characteristic.handle,
            )

        self._characteristics[characteristic.handle] = characteristic

    def get_characteristic(
        self, uuid: Union[str, UUID]
    ) -> Union[BleakGATTCharacteristic, None]:
        """Get a characteristic by UUID.

        Args:
            uuid: The UUID to match.

        Returns:
            The first characteristic matching ``uuid`` or ``None`` if no
            matching characteristic was found.
        """
        uuid = normalize_uuid_str(str(uuid))

        try:
            return next(
                filter(lambda x: x.uuid == uuid, self._characteristics.values())
            )
        except StopIteration:
            return None


class BleakGATTServiceCollection:
    """Simple data container for storing the peripheral's service complement."""

    def __init__(self) -> None:
        self.__services: dict[int, BleakGATTService] = {}
        self.__characteristics: dict[int, BleakGATTCharacteristic] = {}
        self.__descriptors: dict[int, BleakGATTDescriptor] = {}

    def __getitem__(
        self, item: Union[str, int, UUID]
    ) -> Optional[
        Union[BleakGATTService, BleakGATTCharacteristic, BleakGATTDescriptor]
    ]:
        """Get a service, characteristic or descriptor from uuid or handle"""
        return (
            self.get_service(item)
            or self.get_characteristic(item)
            or self.get_descriptor(cast(int, item))
        )

    def __iter__(self) -> Iterator[BleakGATTService]:
        """Returns an iterator over all BleakGATTService objects"""
        return iter(self.services.values())

    @property
    def services(self) -> dict[int, BleakGATTService]:
        """Returns dictionary of handles mapping to BleakGATTService"""
        return self.__services

    @property
    def characteristics(self) -> dict[int, BleakGATTCharacteristic]:
        """Returns dictionary of handles mapping to BleakGATTCharacteristic"""
        return self.__characteristics

    @property
    def descriptors(self) -> dict[int, BleakGATTDescriptor]:
        """Returns a dictionary of integer handles mapping to BleakGATTDescriptor"""
        return self.__descriptors

    def add_service(self, service: BleakGATTService) -> None:
        """Add a :py:class:`~BleakGATTService` to the service collection.

        Should not be used by end user, but rather by `bleak` itself.
        """
        if service.handle not in self.__services:
            self.__services[service.handle] = service
        else:
            logger.error(
                "The service '%s' is already present in this BleakGATTServiceCollection!",
                service.handle,
            )

    def get_service(
        self, specifier: Union[int, str, UUID]
    ) -> Optional[BleakGATTService]:
        """Get a service by handle (int) or UUID (str or uuid.UUID)"""
        if isinstance(specifier, int):
            return self.services.get(specifier)

        uuid = normalize_uuid_str(str(specifier))

        x = list(
            filter(
                lambda x: x.uuid == uuid,
                self.services.values(),
            )
        )

        if len(x) > 1:
            raise BleakError(
                "Multiple Services with this UUID, refer to your desired service by the `handle` attribute instead."
            )

        return x[0] if x else None

    def add_characteristic(self, characteristic: BleakGATTCharacteristic) -> None:
        """Add a :py:class:`~BleakGATTCharacteristic` to the service collection.

        Should not be used by end user, but rather by `bleak` itself.
        """
        if characteristic.handle not in self.__characteristics:
            self.__characteristics[characteristic.handle] = characteristic
            self.__services[characteristic.service_handle].add_characteristic(
                characteristic
            )
        else:
            logger.error(
                "The characteristic '%s' is already present in this BleakGATTServiceCollection!",
                characteristic.handle,
            )

    def get_characteristic(
        self, specifier: Union[int, str, UUID]
    ) -> Optional[BleakGATTCharacteristic]:
        """Get a characteristic by handle (int) or UUID (str or uuid.UUID)"""
        if isinstance(specifier, int):
            return self.characteristics.get(specifier)

        uuid = normalize_uuid_str(str(specifier))

        # Assume uuid usage.
        x = list(
            filter(
                lambda x: x.uuid == uuid,
                self.characteristics.values(),
            )
        )

        if len(x) > 1:
            raise BleakError(
                "Multiple Characteristics with this UUID, refer to your desired characteristic by the `handle` attribute instead."
            )

        return x[0] if x else None

    def add_descriptor(self, descriptor: BleakGATTDescriptor) -> None:
        """Add a :py:class:`~BleakGATTDescriptor` to the service collection.

        Should not be used by end user, but rather by `bleak` itself.
        """
        if descriptor.handle not in self.__descriptors:
            self.__descriptors[descriptor.handle] = descriptor
            self.__characteristics[descriptor.characteristic_handle].add_descriptor(
                descriptor
            )
        else:
            logger.error(
                "The descriptor '%s' is already present in this BleakGATTServiceCollection!",
                descriptor.handle,
            )

    def get_descriptor(self, handle: int) -> Optional[BleakGATTDescriptor]:
        """Get a descriptor by integer handle"""
        return self.descriptors.get(handle)



# -*- coding: utf-8 -*-
# Created on 2020-08-19 by hbldh <henrik.blidh@nedomkull.com>
"""
BLE Client for Windows 10 systems, implemented with WinRT.
"""
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "win32":
        assert False, "This backend is only available on Windows"

import asyncio
import logging
import uuid
from collections.abc import Callable
from contextvars import Context
from ctypes import WinError
from typing import Any, Generic, Optional, Protocol, Sequence, TypeVar, Union, cast
from warnings import warn

if sys.version_info < (3, 12):
    from typing_extensions import Buffer, override
else:
    from collections.abc import Buffer
    from typing import override

if sys.version_info < (3, 11):
    from async_timeout import timeout as async_timeout
    from typing_extensions import Self, assert_never
else:
    from asyncio import timeout as async_timeout
    from typing import Self, assert_never

from winrt.system import Object
from winrt.windows.devices.bluetooth import (
    BluetoothAddressType,
    BluetoothCacheMode,
    BluetoothError,
    BluetoothLEDevice,
)
from winrt.windows.devices.bluetooth.genericattributeprofile import (
    GattCharacteristic,
    GattCharacteristicProperties,
    GattClientCharacteristicConfigurationDescriptorValue,
    GattCommunicationStatus,
    GattDescriptor,
    GattDeviceService,
    GattSession,
    GattSessionStatus,
    GattSessionStatusChangedEventArgs,
    GattValueChangedEventArgs,
    GattWriteOption,
)
from winrt.windows.devices.enumeration import (
    DeviceInformation,
    DeviceInformationCustomPairing,
    DevicePairingKinds,
    DevicePairingProtectionLevel,
    DevicePairingRequestedEventArgs,
    DevicePairingResultStatus,
    DeviceUnpairingResultStatus,
)
from winrt.windows.foundation import (
    AsyncStatus,
    EventRegistrationToken,
    IAsyncOperation,
)
from winrt.windows.storage.streams import Buffer as WinBuffer

from bleak import BleakScanner
from bleak.args.winrt import WinRTClientArgs as _WinRTClientArgs
from bleak.assigned_numbers import gatt_char_props_to_strs
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.client import BaseBleakClient, NotifyCallback
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.backends.winrt.scanner import BleakScannerWinRT, RawAdvData
from bleak.exc import PROTOCOL_ERROR_CODES, BleakDeviceNotFoundError, BleakError

logger = logging.getLogger(__name__)


def __getattr__(name: str):
    if name == "WinRTClientArgs":
        warn(
            "importing WinRTClientArgs from bleak.backends.winrt.client is deprecated, use bleak.args.winrt instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return _WinRTClientArgs
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class _Result(Protocol):
    @property
    def status(self) -> GattCommunicationStatus: ...
    @property
    def protocol_error(self) -> Optional[int]: ...


def _address_to_int(address: str) -> int:
    """Converts the Bluetooth device address string to its representing integer

    Args:
        address (str): Bluetooth device address to convert

    Returns:
        int: integer representation of the given Bluetooth device address
    """
    _address_separators = [":", "-"]
    for char in _address_separators:
        address = address.replace(char, "")

    return int(address, base=16)


def _ensure_success(result: _Result, attr: Optional[str], fail_msg: str) -> Any:
    """
    Ensures that *status* is ``GattCommunicationStatus.SUCCESS``, otherwise
    raises ``BleakError``.

    Args:
        result: The result returned by a WinRT API method.
        attr: The name of the attribute containing the result.
        fail_msg: A message to include in the exception.
    """
    status = result.status if hasattr(result, "status") else result

    if status == GattCommunicationStatus.SUCCESS:
        return None if attr is None else getattr(result, attr)

    if status == GattCommunicationStatus.PROTOCOL_ERROR:
        assert result.protocol_error is not None
        err = PROTOCOL_ERROR_CODES.get(result.protocol_error, "Unknown")
        raise BleakError(
            f"{fail_msg}: Protocol Error 0x{result.protocol_error:02X}: {err}"
        )

    if status == GattCommunicationStatus.ACCESS_DENIED:
        raise BleakError(f"{fail_msg}: Access Denied")

    if status == GattCommunicationStatus.UNREACHABLE:
        raise BleakError(f"{fail_msg}: Unreachable")

    raise BleakError(f"{fail_msg}: Unexpected status code 0x{status:02X}")


class BleakClientWinRT(BaseBleakClient):
    """Native Windows Bleak Client.

    Args:
        address_or_ble_device (str or BLEDevice): The Bluetooth address of the BLE peripheral
            to connect to or the ``BLEDevice`` object representing it.
        services: Optional set of service UUIDs that will be used.
        winrt (dict): A dictionary of Windows-specific configuration values.
        **timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.
    """

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        services: Optional[set[str]] = None,
        *,
        winrt: _WinRTClientArgs,
        **kwargs: Any,
    ):
        super(BleakClientWinRT, self).__init__(address_or_ble_device, **kwargs)

        # Backend specific. WinRT objects.
        if isinstance(address_or_ble_device, BLEDevice):
            data: RawAdvData = address_or_ble_device.details
            args = data.adv or data.scan
            assert args
            self._device_info = args.bluetooth_address
        else:
            self._device_info = None

        self._requested_services = (
            [uuid.UUID(s) for s in services] if services else None
        )
        self._requester: Optional[BluetoothLEDevice] = None
        self._services_changed_events: list[asyncio.Event] = []
        self._session_active_events: list[asyncio.Event] = []
        self._session: Optional[GattSession] = None
        self._notification_callbacks: dict[int, EventRegistrationToken] = {}

        # os-specific options
        self._use_cached_services = winrt.get("use_cached_services")
        self._address_type = winrt.get("address_type")
        self._retry_on_services_changed = False

        self._session_services_changed_token: Optional[EventRegistrationToken] = None
        self._session_status_changed_token: Optional[EventRegistrationToken] = None
        self._max_pdu_size_changed_token: Optional[EventRegistrationToken] = None

    def __str__(self) -> str:
        return f"{type(self).__name__} ({self.address})"

    # Connectivity methods

    async def _create_requester(self, bluetooth_address: int) -> BluetoothLEDevice:
        if self._address_type is not None:
            requester = await BluetoothLEDevice.from_bluetooth_address_with_bluetooth_address_type_async(
                bluetooth_address,
                (
                    BluetoothAddressType.PUBLIC
                    if self._address_type == "public"
                    else BluetoothAddressType.RANDOM
                ),
            )
        else:
            requester = await BluetoothLEDevice.from_bluetooth_address_async(
                bluetooth_address
            )

        # https://github.com/microsoft/Windows-universal-samples/issues/1089#issuecomment-487586755
        if requester is None:
            raise BleakDeviceNotFoundError(
                self.address, f"Device with address {self.address} was not found."
            )
        return requester

    @override
    async def connect(self, pair: bool, **kwargs: Any) -> None:
        """Connect to the specified GATT server.

        Keyword Args:
            timeout (float): Timeout for required ``BleakScanner.find_device_by_address`` call. Defaults to 10.0.
        """
        # Try to find the desired device.
        timeout = kwargs.get("timeout", self._timeout)
        if self._device_info is None:
            device = await BleakScanner.find_device_by_address(
                self.address, timeout=timeout, backend=BleakScannerWinRT
            )

            if device is None:
                raise BleakDeviceNotFoundError(
                    self.address, f"Device with address {self.address} was not found."
                )

            data: RawAdvData = device.details
            args = data.adv or data.scan
            assert args
            self._device_info = args.bluetooth_address

        logger.debug("Connecting to BLE device @ %s", self.address)

        loop = asyncio.get_running_loop()

        self._requester = await self._create_requester(self._device_info)

        if pair:
            await self.pair(**kwargs)

        def handle_services_changed() -> None:
            if not self._services_changed_events:
                logger.warning("%s: unhandled services changed event", self.address)
            else:
                for event in self._services_changed_events:
                    event.set()

        def services_changed_handler(sender: BluetoothLEDevice, args: Object) -> None:
            logger.debug("%s: services changed", self.address)
            loop.call_soon_threadsafe(handle_services_changed)

        self._services_changed_token = self._requester.add_gatt_services_changed(
            services_changed_handler
        )

        # Called on disconnect event or on failure to connect.
        def handle_disconnect() -> None:
            if self._requester:
                if self._services_changed_token:
                    self._requester.remove_gatt_services_changed(
                        self._services_changed_token
                    )
                    self._services_changed_token = None

                logger.debug("closing requester")
                self._requester.close()
                self._requester = None

            if self._session:
                if self._session_status_changed_token:
                    self._session.remove_session_status_changed(
                        self._session_status_changed_token
                    )
                    self._session_status_changed_token = None

                if self._max_pdu_size_changed_token:
                    self._session.remove_max_pdu_size_changed(
                        self._max_pdu_size_changed_token
                    )
                    self._max_pdu_size_changed_token = None

                logger.debug("closing session")
                self._session.close()
                self._session = None

        is_connect_complete = False

        def handle_session_status_changed(
            args: GattSessionStatusChangedEventArgs,
        ) -> None:
            if args.error != BluetoothError.SUCCESS:
                logger.error("Unhandled GATT error %r", args.error)

            if args.status == GattSessionStatus.ACTIVE:
                for e in self._session_active_events:
                    e.set()

            # Don't run this if we have not exited from the connect method yet.
            # Cleanup is handled by the connect method in that case.
            elif args.status == GattSessionStatus.CLOSED and is_connect_complete:
                if self._disconnected_callback:
                    self._disconnected_callback()

                handle_disconnect()

        # this is the WinRT event handler will be called on another thread
        def session_status_changed_event_handler(
            sender: GattSession, args: GattSessionStatusChangedEventArgs
        ):
            logger.debug(
                "session_status_changed_event_handler: id: %s, error: %r, status: %r",
                sender.device_id.id,
                args.error,
                args.status,
            )
            loop.call_soon_threadsafe(handle_session_status_changed, args)

        def max_pdu_size_changed_handler(sender: GattSession, args: Object) -> None:
            try:
                max_pdu_size = sender.max_pdu_size
            except OSError:
                # There is a race condition where this event was already
                # queued when the GattSession object was closed. In that
                # case, we get a Windows error which we can just ignore.
                return

            logger.debug("max_pdu_size_changed_handler: %d", max_pdu_size)

        # Start a GATT Session to connect
        event = asyncio.Event()
        self._session_active_events.append(event)
        try:
            self._session = await GattSession.from_device_id_async(
                self._requester.bluetooth_device_id
            )

            if not self._session.can_maintain_connection:
                raise BleakError("device does not support GATT sessions")

            self._session_status_changed_token = (
                self._session.add_session_status_changed(
                    session_status_changed_event_handler
                )
            )

            self._max_pdu_size_changed_token = self._session.add_max_pdu_size_changed(
                max_pdu_size_changed_handler
            )

            services_changed_event = asyncio.Event()
            self._services_changed_events.append(services_changed_event)

            # If the session is already active, we need to set the event since
            # the session_status_changed event won't fire. This happens, e.g.,
            # when pairing before connecting which causes the device to already
            # be connected.
            if self._session.session_status == GattSessionStatus.ACTIVE:
                event.set()

            try:
                # Windows does not support explicitly connecting to a device.
                # Instead it has the concept of a GATT session that is owned
                # by the calling program.
                self._session.maintain_connection = True
                # This keeps the device connected until we set maintain_connection = False.

                cache_mode = None

                if self._use_cached_services is not None:
                    cache_mode = (
                        BluetoothCacheMode.CACHED
                        if self._use_cached_services
                        else BluetoothCacheMode.UNCACHED
                    )

                # if we receive a services changed event before get_gatt_services_async()
                # finishes, we need to call it again with BluetoothCacheMode.CACHED
                # to ensure we have the correct services as described in
                # https://learn.microsoft.com/en-us/uwp/api/windows.devices.bluetooth.bluetoothledevice.gattserviceschanged
                service_cache_mode = cache_mode

                async with async_timeout(timeout):
                    if self._retry_on_services_changed:
                        while True:
                            services_changed_event.clear()
                            services_changed_event_task = asyncio.create_task(
                                services_changed_event.wait()
                            )

                            get_services_task = asyncio.create_task(
                                self._get_services(
                                    service_cache_mode=service_cache_mode,
                                    cache_mode=cache_mode,
                                )
                            )

                            _, pending = await asyncio.wait(
                                [services_changed_event_task, get_services_task],
                                return_when=asyncio.FIRST_COMPLETED,
                            )

                            for p in pending:
                                p.cancel()

                            if not services_changed_event.is_set():
                                # services did not change while getting services,
                                # so this is the final result
                                self.services = get_services_task.result()
                                break

                            logger.debug(
                                "%s: restarting get services due to services changed event",
                                self.address,
                            )
                            service_cache_mode = BluetoothCacheMode.CACHED

                            # ensure the task ran to completion to avoid OSError
                            # on next call to get_services()
                            try:
                                await get_services_task
                            except OSError:
                                pass
                            except asyncio.CancelledError:
                                pass
                    else:
                        self.services = await self._get_services(
                            service_cache_mode=service_cache_mode,
                            cache_mode=cache_mode,
                        )

                    # a connection may not be made until we request info from the
                    # device, so we have to get services before the GATT session
                    # is set to active
                    await event.wait()

                is_connect_complete = True
            finally:
                self._services_changed_events.remove(services_changed_event)

        except BaseException:
            handle_disconnect()
            raise
        finally:
            self._session_active_events.remove(event)

    @override
    async def disconnect(self) -> None:
        """Disconnect from the specified GATT server."""
        logger.debug("Disconnecting from BLE device...")

        if self.services is None:
            # No connection exists. Either one hasn't been created or
            # we have already called disconnect and closed the gatt
            # connection.
            logger.debug("already disconnected")
            return

        # Remove notifications.
        for handle, event_handler_token in list(self._notification_callbacks.items()):
            char = self.services.get_characteristic(handle)
            assert char
            gatt_char = cast(GattCharacteristic, char.obj)
            gatt_char.remove_value_changed(event_handler_token)
        self._notification_callbacks.clear()

        # Since we can't wait for the session close event (it may never come)
        # we need to synthesize the disconnect event
        if self._session and self._session_status_changed_token:
            self._session.remove_session_status_changed(
                self._session_status_changed_token
            )
            self._session_status_changed_token = None

            if self._disconnected_callback:
                self._disconnected_callback()

        # Dispose all service components that we have requested and created.
        if self.services:
            # HACK: sometimes GattDeviceService.Close() hangs forever, so we
            # add a delay to give the Windows Bluetooth stack some time to
            # "settle" before closing the services
            await asyncio.sleep(0.1)

            for service in self.services:
                service.obj.close()
            self.services = None

        if self._session:
            self._session.close()
            self._session = None

        if self._requester:
            self._requester.close()
            self._requester = None

    @property
    @override
    def is_connected(self) -> bool:
        """Check connection status between this client and the server.

        Returns:
            Boolean representing connection status.

        """
        return (
            False
            if self._session is None
            else self._session.session_status == GattSessionStatus.ACTIVE
        )

    @property
    @override
    def name(self) -> str:
        """See :meth:`bleak.BleakClient.name`."""
        if self._requester is None:
            raise BleakError("Not connected")
        return self._requester.name

    @property
    @override
    def mtu_size(self) -> int:
        """Get ATT MTU size for active connection"""
        return self._session.max_pdu_size

    @override
    async def pair(
        self,
        **kwargs: Any,
    ) -> None:
        """Attempts to pair with the device.

        Keyword Args:
            protection_level (int): A ``DevicePairingProtectionLevel`` enum value:

                1. None - Pair the device using no levels of protection.
                2. Encryption - Pair the device using encryption.
                3. EncryptionAndAuthentication - Pair the device using
                   encryption and authentication.

                .. versionchanged:: 1.0
                    Issues :class:`DeprecationWarning` if used. The default
                    behavior has changed and this argument should no longer
                    be needed.
        """
        assert self._requester

        # New local device information object created since the object from the requester isn't updated
        device_information = await DeviceInformation.create_from_id_async(
            self._requester.device_information.id
        )

        if device_information.pairing.is_paired:
            logging.debug("Device is already paired. Skipping pairing.")
            return

        if not device_information.pairing.can_pair:
            raise BleakError("Device does not support pairing")

        protection_level = kwargs.get("protection_level")

        # Currently only supporting Just Works solutions...
        ceremony = DevicePairingKinds.CONFIRM_ONLY
        custom_pairing = device_information.pairing.custom

        def handler(
            sender: DeviceInformationCustomPairing,
            args: DevicePairingRequestedEventArgs,
        ):
            args.accept()

        pairing_requested_token = custom_pairing.add_pairing_requested(handler)

        try:
            if protection_level is not None:
                warn(
                    "protection_level is deprecated and will be removed in a future version. The default protection level has changed, so it should be safe to omit this argument.",
                    DeprecationWarning,
                    2,
                )
                pairing_result = await custom_pairing.pair_with_protection_level_async(
                    ceremony, protection_level
                )
            else:
                for level in (
                    DevicePairingProtectionLevel.ENCRYPTION_AND_AUTHENTICATION,
                    DevicePairingProtectionLevel.ENCRYPTION,
                ):
                    pairing_result = (
                        await custom_pairing.pair_with_protection_level_async(
                            ceremony, level
                        )
                    )
                    if (
                        pairing_result.status
                        != DevicePairingResultStatus.PROTECTION_LEVEL_COULD_NOT_BE_MET
                    ):
                        break

                    logger.debug("Protection level %r not met. Retrying.", level)
                else:
                    pairing_result = await custom_pairing.pair_async(ceremony)

        except Exception as e:
            raise BleakError("Failure trying to pair with device!") from e
        finally:
            custom_pairing.remove_pairing_requested(pairing_requested_token)

        if pairing_result.status not in (
            DevicePairingResultStatus.PAIRED,
            DevicePairingResultStatus.ALREADY_PAIRED,
        ):
            raise BleakError(
                f"Could not pair with device: {pairing_result.status.name}"
            )

        if logger.isEnabledFor(logging.DEBUG):
            # pairing_result.protection_level_used doesn't seem to return
            # accurate information if we don't update the DeviceInformation
            # first.
            device_information = await DeviceInformation.create_from_id_async(
                self._requester.device_information.id
            )

            logger.debug(
                "Paired to device with protection level %s.",
                pairing_result.protection_level_used.name,
            )

    @override
    async def unpair(self) -> None:
        """Attempts to unpair from the device.

        N.B. unpairing also leads to disconnection in the Windows backend.
        """
        device = await self._create_requester(
            self._device_info
            if self._device_info is not None
            else _address_to_int(self.address)
        )

        try:
            unpairing_result = await device.device_information.pairing.unpair_async()
            if unpairing_result.status not in (
                DeviceUnpairingResultStatus.UNPAIRED,
                DeviceUnpairingResultStatus.ALREADY_UNPAIRED,
            ):
                raise BleakError(
                    f"Could not unpair with device: {unpairing_result.status}"
                )
            logger.info("Unpaired with device.")
        finally:
            device.close()

    # GATT services methods

    async def _get_services(
        self,
        *,
        service_cache_mode: Optional[BluetoothCacheMode] = None,
        cache_mode: Optional[BluetoothCacheMode] = None,
        **kwargs: Any,
    ) -> BleakGATTServiceCollection:
        """Get all services registered for this GATT server.

        Returns:
           A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.

        """

        # Return the Service Collection.
        if self.services is not None:
            return self.services

        logger.debug(
            "getting services (service_cache_mode=%r, cache_mode=%r)...",
            service_cache_mode,
            cache_mode,
        )

        new_services = BleakGATTServiceCollection()
        services: Sequence[GattDeviceService]

        assert self._requester

        if self._requested_services is None:
            if service_cache_mode is not None:
                result = await FutureLike(
                    self._requester.get_gatt_services_with_cache_mode_async(
                        service_cache_mode
                    )
                )
            else:
                result = await FutureLike(self._requester.get_gatt_services_async())

            services = _ensure_success(
                result,
                "services",
                "Could not get GATT services",
            )
        else:
            services = []
            # REVISIT: should properly dispose services on cancel or protect from cancellation

            for s in self._requested_services:
                if service_cache_mode is not None:
                    result = await FutureLike(
                        self._requester.get_gatt_services_for_uuid_with_cache_mode_async(
                            s, service_cache_mode
                        )
                    )
                else:
                    result = await FutureLike(
                        self._requester.get_gatt_services_for_uuid_async(s)
                    )

                services.extend(
                    _ensure_success(
                        result,
                        "services",
                        "Could not get GATT services",
                    )
                )

        try:
            for service in services:
                if cache_mode is not None:
                    result = await FutureLike(
                        service.get_characteristics_with_cache_mode_async(cache_mode)
                    )
                else:
                    result = await FutureLike(service.get_characteristics_async())

                if result.status == GattCommunicationStatus.ACCESS_DENIED:
                    # Windows does not allow access to services "owned" by the
                    # OS. This includes services like HID and Bond Manager.
                    logger.debug(
                        "skipping service %s due to access denied", service.uuid
                    )
                    continue

                characteristics: Sequence[GattCharacteristic] = _ensure_success(
                    result,
                    "characteristics",
                    f"Could not get GATT characteristics for service {service.uuid} ({service.attribute_handle})",
                )

                serv = BleakGATTService(
                    service, service.attribute_handle, str(service.uuid)
                )
                new_services.add_service(serv)

                for characteristic in characteristics:
                    if cache_mode is not None:
                        result = await FutureLike(
                            characteristic.get_descriptors_with_cache_mode_async(
                                cache_mode
                            )
                        )
                    else:
                        result = await FutureLike(
                            characteristic.get_descriptors_async()
                        )

                    descriptors: Sequence[GattDescriptor] = _ensure_success(
                        result,
                        "descriptors",
                        f"Could not get GATT descriptors for characteristic {characteristic.uuid} ({characteristic.attribute_handle})",
                    )

                    char = BleakGATTCharacteristic(
                        characteristic,
                        characteristic.attribute_handle,
                        str(characteristic.uuid),
                        list(
                            gatt_char_props_to_strs(
                                characteristic.characteristic_properties
                            )
                        ),
                        lambda: self._session.max_pdu_size - 3,
                        serv,
                    )

                    new_services.add_characteristic(char)

                    for descriptor in descriptors:
                        desc = BleakGATTDescriptor(
                            descriptor,
                            descriptor.attribute_handle,
                            str(descriptor.uuid),
                            char,
                        )
                        new_services.add_descriptor(desc)

            return new_services
        except BaseException:
            # Don't leak services. WinRT is quite particular about services
            # being closed.
            logger.debug("disposing service objects")

            # HACK: sometimes GattDeviceService.Close() hangs forever, so we
            # add a delay to give the Windows Bluetooth stack some time to
            # "settle" before closing the services
            await asyncio.sleep(0.1)

            for service in services:
                service.close()
            raise

    # I/O methods

    @override
    async def read_gatt_char(
        self, characteristic: BleakGATTCharacteristic, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to read from.

        Keyword Args:
            use_cached (bool): ``False`` forces Windows to read the value from the
                device again and not use its own cached value. Defaults to ``False``.

        Returns:
            (bytearray) The read data.

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        assert self.services

        use_cached = kwargs.get("use_cached", False)

        gatt_char = cast(GattCharacteristic, characteristic.obj)

        value = bytearray(
            _ensure_success(
                await gatt_char.read_value_with_cache_mode_async(
                    BluetoothCacheMode.CACHED
                    if use_cached
                    else BluetoothCacheMode.UNCACHED
                ),
                "value",
                f"Could not read characteristic handle {characteristic.handle}",
            )
        )

        logger.debug("Read Characteristic %04X : %s", characteristic.handle, value)

        return value

    @override
    async def read_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.

        Keyword Args:
            use_cached (bool): `False` forces Windows to read the value from the
                device again and not use its own cached value. Defaults to `False`.

        Returns:
            The read data.

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        assert self.services

        use_cached = kwargs.get("use_cached", False)
        gatt_desc = cast(GattDescriptor, descriptor.obj)

        value = bytearray(
            _ensure_success(
                await gatt_desc.read_value_with_cache_mode_async(
                    BluetoothCacheMode.CACHED
                    if use_cached
                    else BluetoothCacheMode.UNCACHED
                ),
                "value",
                f"Could not read Descriptor value for {descriptor.handle:04X}",
            )
        )

        logger.debug("Read Descriptor %04X : %s", descriptor.handle, value)

        return value

    @override
    async def write_gatt_char(
        self, characteristic: BleakGATTCharacteristic, data: Buffer, response: bool
    ) -> None:
        if not self.is_connected:
            raise BleakError("Not connected")

        buf = WinBuffer(len(data))
        buf.length = buf.capacity

        with memoryview(buf) as mv:
            mv[:] = data

        gatt_char = cast(GattCharacteristic, characteristic.obj)

        _ensure_success(
            await gatt_char.write_value_with_result_and_option_async(
                buf,
                (
                    GattWriteOption.WRITE_WITH_RESPONSE
                    if response
                    else GattWriteOption.WRITE_WITHOUT_RESPONSE
                ),
            ),
            None,
            f"Could not write value {data} to characteristic {characteristic.handle:04X}",
        )

    @override
    async def write_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, data: Buffer
    ) -> None:
        """Perform a write operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.
            data: The data to send (any bytes-like object).

        """
        if not self.is_connected:
            raise BleakError("Not connected")

        assert self.services

        buf = WinBuffer(len(data))
        buf.length = buf.capacity

        with memoryview(buf) as mv:
            mv[:] = data

        gatt_desc = cast(GattDescriptor, descriptor.obj)

        _ensure_success(
            await gatt_desc.write_value_with_result_async(buf),
            None,
            f"Could not write value {data!r} to descriptor {descriptor.handle:04X}",
        )

        logger.debug("Write Descriptor %04X : %s", descriptor.handle, data)

    @override
    async def start_notify(
        self,
        characteristic: BleakGATTCharacteristic,
        callback: NotifyCallback,
        **kwargs: Any,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.

        Keyword Args:
            force_indicate (bool): If this is set to True, then Bleak will set up a indication request instead of a
                notification request, given that the characteristic supports notifications as well as indications.
        """
        winrt_char = cast(GattCharacteristic, characteristic.obj)

        # If we want to force indicate even when notify is available, also check if the device
        # actually supports indicate as well.
        if not kwargs.get("force_indicate", False) and (
            winrt_char.characteristic_properties & GattCharacteristicProperties.NOTIFY
        ):
            cccd = GattClientCharacteristicConfigurationDescriptorValue.NOTIFY
        elif (
            winrt_char.characteristic_properties & GattCharacteristicProperties.INDICATE
        ):
            cccd = GattClientCharacteristicConfigurationDescriptorValue.INDICATE
        else:
            raise BleakError(
                "characteristic does not support notifications or indications"
            )

        loop = asyncio.get_running_loop()

        def handle_value_changed(
            sender: GattCharacteristic, args: GattValueChangedEventArgs
        ) -> None:
            value = bytearray(args.characteristic_value)
            loop.call_soon_threadsafe(callback, value)

        event_handler_token = winrt_char.add_value_changed(handle_value_changed)
        self._notification_callbacks[characteristic.handle] = event_handler_token

        try:
            _ensure_success(
                await winrt_char.write_client_characteristic_configuration_descriptor_with_result_async(
                    cccd
                ),
                None,
                f"Could not start notify on {characteristic.handle:04X}",
            )
        except BaseException:
            # This usually happens when a device reports that it supports indicate,
            # but it actually doesn't.
            if characteristic.handle in self._notification_callbacks:
                event_handler_token = self._notification_callbacks.pop(
                    characteristic.handle
                )
                winrt_char.remove_value_changed(event_handler_token)

            raise

    @override
    async def stop_notify(self, characteristic: BleakGATTCharacteristic) -> None:
        """Deactivate notification/indication on a specified characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to deactivate
                notification/indication on.
        """
        if not self.is_connected:
            raise BleakError("Not connected")

        assert self.services

        gatt_char = cast(GattCharacteristic, characteristic.obj)

        _ensure_success(
            await gatt_char.write_client_characteristic_configuration_descriptor_with_result_async(
                GattClientCharacteristicConfigurationDescriptorValue.NONE
            ),
            None,
            f"Could not stop notify on {characteristic.handle:04X}",
        )

        event_handler_token = self._notification_callbacks.pop(characteristic.handle)
        gatt_char.remove_value_changed(event_handler_token)


T = TypeVar("T")


class FutureLike(Generic[T]):
    """
    Wraps a WinRT IAsyncOperation in a "future-like" object so that it can
    be passed to Python APIs.

    Needed until https://github.com/pywinrt/pywinrt/issues/14
    """

    _asyncio_future_blocking = False

    def __init__(self: Self, op: IAsyncOperation[T]) -> None:
        self._op = op
        self._callbacks: list[Callable[[Self], None]] = []
        self._loop = asyncio.get_running_loop()
        self._cancel_requested = False
        self._result = None

        def call_callbacks() -> None:
            for c in self._callbacks:
                c(self)

        def call_callbacks_threadsafe(
            op: IAsyncOperation[T], status: AsyncStatus
        ) -> None:
            if status == AsyncStatus.COMPLETED:
                # have to get result on this thread, otherwise it may not return correct value
                self._result = op.get_results()

            self._loop.call_soon_threadsafe(call_callbacks)

        op.completed = call_callbacks_threadsafe

    def result(self) -> T:
        if self._op.status == AsyncStatus.STARTED:
            raise asyncio.InvalidStateError

        if self._op.status == AsyncStatus.COMPLETED:
            if self._cancel_requested:
                raise asyncio.CancelledError

            assert self._result

            return self._result

        if self._op.status == AsyncStatus.CANCELED:
            raise asyncio.CancelledError

        if self._op.status == AsyncStatus.ERROR:
            if self._cancel_requested:
                raise asyncio.CancelledError

            error_code = self._op.error_code.value
            raise WinError(error_code)

        assert_never(self._op.status)

    def done(self) -> bool:
        return self._op.status != AsyncStatus.STARTED

    def cancelled(self) -> bool:
        return self._cancel_requested or self._op.status == AsyncStatus.CANCELED

    def add_done_callback(
        self,
        callback: Callable[[Self], None],
        *,
        context: Optional[Context] = None,
    ) -> None:
        self._callbacks.append(callback)

    def remove_done_callback(self, callback: Callable[[Self], None]) -> None:
        self._callbacks.remove(callback)

    def cancel(self, msg: Optional[str] = None) -> bool:
        if self._cancel_requested or self._op.status != AsyncStatus.STARTED:
            return False

        self._cancel_requested = True
        self._op.cancel()

        return True

    def exception(self) -> Optional[Exception]:
        if self._op.status == AsyncStatus.STARTED:
            raise asyncio.InvalidStateError

        if self._op.status == AsyncStatus.COMPLETED:
            if self._cancel_requested:
                raise asyncio.CancelledError

            return None

        if self._op.status == AsyncStatus.CANCELED:
            raise asyncio.CancelledError

        if self._op.status == AsyncStatus.ERROR:
            if self._cancel_requested:
                raise asyncio.CancelledError

            error_code = self._op.error_code.value

            return WinError(error_code)

        assert_never(self._op.status)

    def get_loop(self) -> asyncio.AbstractEventLoop:
        return self._loop

    def __await__(self):
        if not self.done():
            self._asyncio_future_blocking = True
            yield self  # This tells Task to wait for completion.

        if not self.done():
            raise RuntimeError("await wasn't used with future")

        return self.result()  # May raise too.



"""
Advertisement Monitor
---------------------

This module contains types associated with the BlueZ D-Bus `advertisement
monitor api <https://github.com/bluez/bluez/blob/master/doc/org.bluez.AdvertisementMonitor.rst>`.
"""

import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "linux":
        assert False, "This backend is only available on Linux"

import logging
from collections.abc import Iterable
from typing import Any, no_type_check
from warnings import warn

from dbus_fast import PropertyAccess
from dbus_fast.service import ServiceInterface, dbus_property, method

from bleak.args.bluez import OrPattern as _OrPattern
from bleak.args.bluez import OrPatternLike as _OrPatternLike
from bleak.backends.bluezdbus import defs

logger = logging.getLogger(__name__)

_DEPRECATED: dict[str, Any] = {
    "OrPattern": _OrPattern,
    "OrPatternLike": _OrPatternLike,
}


def __getattr__(name: str):
    if value := _DEPRECATED.get(name):
        warn(
            f"importing {name} from bleak.backends.bluezdbus.advertisement_monitor is deprecated, use bleak.args.bluez instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return value
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


class AdvertisementMonitor(ServiceInterface):
    """
    Implementation of the org.bluez.AdvertisementMonitor1 D-Bus interface.

    The BlueZ advertisement monitor API design seems to be just for device
    presence (is it in range or out of range), but this isn't really what
    we want in Bleak, we want to monitor changes in advertisement data, just
    like in active scanning.

    So the only thing we are using here is the "or_patterns" since it is
    currently required, but really we don't need that either. Hopefully an
    "all" "Type" could be added to BlueZ in the future.
    """

    def __init__(
        self,
        or_patterns: Iterable[_OrPatternLike],
    ):
        """
        Args:
            or_patterns:
                List of or patterns that will be returned by the ``Patterns`` property.
        """
        super().__init__(defs.ADVERTISEMENT_MONITOR_INTERFACE)
        # dbus_fast marshaling requires list instead of tuple
        self._or_patterns = [list(p) for p in or_patterns]

    @method()
    def Release(self):
        logger.debug("Release")

    @method()
    def Activate(self):
        logger.debug("Activate")

    # REVISIT: mypy is broke, so we have to add redundant @no_type_check
    # https://github.com/python/mypy/issues/6583

    @method()
    @no_type_check
    def DeviceFound(self, device: "o"):  # noqa: F821
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("DeviceFound %s", device)

    @method()
    @no_type_check
    def DeviceLost(self, device: "o"):  # noqa: F821
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("DeviceLost %s", device)

    @dbus_property(PropertyAccess.READ)
    @no_type_check
    def Type(self) -> "s":  # noqa: F821
        # this is currently the only type supported in BlueZ
        return "or_patterns"

    @dbus_property(PropertyAccess.READ, disabled=True)
    @no_type_check
    def RSSILowThreshold(self) -> "n":  # noqa: F821
        ...

    @dbus_property(PropertyAccess.READ, disabled=True)
    @no_type_check
    def RSSIHighThreshold(self) -> "n":  # noqa: F821
        ...

    @dbus_property(PropertyAccess.READ, disabled=True)
    @no_type_check
    def RSSILowTimeout(self) -> "q":  # noqa: F821
        ...

    @dbus_property(PropertyAccess.READ, disabled=True)
    @no_type_check
    def RSSIHighTimeout(self) -> "q":  # noqa: F821
        ...

    @dbus_property(PropertyAccess.READ, disabled=True)
    @no_type_check
    def RSSISamplingPeriod(self) -> "q":  # noqa: F821
        ...

    @dbus_property(PropertyAccess.READ)
    @no_type_check
    def Patterns(self) -> "a(yyay)":  # noqa: F821
        return self._or_patterns



# -*- coding: utf-8 -*-
"""
BLE Client for python-for-android
"""
import sys
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    if sys.platform != "android":
        assert False, "This backend is only available on Android"

import asyncio
import logging
import uuid
import warnings
from typing import Any, Optional, Union

if sys.version_info < (3, 12):
    from typing_extensions import override
else:
    from typing import override

from android.broadcast import BroadcastReceiver
from jnius import java_method

from bleak.assigned_numbers import gatt_char_props_to_strs
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.backends.client import BaseBleakClient, NotifyCallback
from bleak.backends.descriptor import BleakGATTDescriptor
from bleak.backends.device import BLEDevice
from bleak.backends.p4android import defs, utils
from bleak.backends.service import BleakGATTService, BleakGATTServiceCollection
from bleak.exc import BleakError

logger = logging.getLogger(__name__)


class BleakClientP4Android(BaseBleakClient):
    """A python-for-android Bleak Client

    Args:
        address_or_ble_device:
            The Bluetooth address of the BLE peripheral to connect to or the
            :class:`BLEDevice` object representing it.
        services:
            Optional set of services UUIDs to filter.
    """

    def __init__(
        self,
        address_or_ble_device: Union[BLEDevice, str],
        services: Optional[set[uuid.UUID]],
        **kwargs,
    ):
        super(BleakClientP4Android, self).__init__(address_or_ble_device, **kwargs)
        self._requested_services = (
            set(map(defs.UUID.fromString, services)) if services else None
        )
        # kwarg "device" is for backwards compatibility
        self.__adapter = kwargs.get("adapter", kwargs.get("device", None))
        self.__gatt = None
        self.__mtu = 23

        self.__callbacks = None

    # Connectivity methods

    @override
    async def connect(self, pair: bool, **kwargs) -> None:
        """Connect to the specified GATT server."""
        if pair:
            logger.warning("Pairing during connect is not implemented on Android")

        loop = asyncio.get_running_loop()

        self.__adapter = defs.BluetoothAdapter.getDefaultAdapter()
        if self.__adapter is None:
            raise BleakError("Bluetooth is not supported on this hardware platform")
        if self.__adapter.getState() != defs.BluetoothAdapter.STATE_ON:
            raise BleakError("Bluetooth is not turned on")

        self.__device = self.__adapter.getRemoteDevice(self.address)

        self.__callbacks = _PythonBluetoothGattCallback(self, loop)

        self._subscriptions = {}

        logger.debug(f"Connecting to BLE device @ {self.address}")

        (self.__gatt,) = await self.__callbacks.perform_and_wait(
            dispatchApi=self.__device.connectGatt,
            dispatchParams=(
                defs.context,
                False,
                self.__callbacks.java,
                defs.BluetoothDevice.TRANSPORT_LE,
            ),
            resultApi="onConnectionStateChange",
            resultExpected=(defs.BluetoothProfile.STATE_CONNECTED,),
            return_indicates_status=False,
        )

        try:
            logger.debug("Connection successful.")

            # unlike other backends, Android doesn't automatically negotiate
            # the MTU, so we request the largest size possible like BlueZ
            logger.debug("requesting mtu...")
            (self.__mtu,) = await self.__callbacks.perform_and_wait(
                dispatchApi=self.__gatt.requestMtu,
                dispatchParams=(517,),
                resultApi="onMtuChanged",
            )

            logger.debug("discovering services...")
            await self.__callbacks.perform_and_wait(
                dispatchApi=self.__gatt.discoverServices,
                dispatchParams=(),
                resultApi="onServicesDiscovered",
            )

            await self._get_services()
        except BaseException:
            # if connecting is canceled or one of the above fails, we need to
            # disconnect
            try:
                await self.disconnect()
            except Exception:
                pass
            raise

    @override
    async def disconnect(self) -> None:
        """Disconnect from the specified GATT server."""
        logger.debug("Disconnecting from BLE device...")
        if self.__gatt is None:
            # No connection exists. Either one hasn't been created or
            # we have already called disconnect and closed the gatt
            # connection.
            logger.debug("already disconnected")
            return

        # Try to disconnect the actual device/peripheral
        try:
            await self.__callbacks.perform_and_wait(
                dispatchApi=self.__gatt.disconnect,
                dispatchParams=(),
                resultApi="onConnectionStateChange",
                resultExpected=(defs.BluetoothProfile.STATE_DISCONNECTED,),
                unless_already=True,
                return_indicates_status=False,
            )
            self.__gatt.close()
        except Exception as e:
            logger.error(f"Attempt to disconnect device failed: {e}")

        self.__gatt = None
        self.__callbacks = None

        # Reset all stored services.
        self.services = None

    @override
    async def pair(self, *args, **kwargs) -> None:
        """Pair with the peripheral.

        You can use ConnectDevice method if you already know the MAC address of the device.
        Else you need to StartDiscovery, Trust, Pair and Connect in sequence.
        """
        loop = asyncio.get_running_loop()

        bondedFuture = loop.create_future()

        def handleBondStateChanged(context, intent):
            bond_state = intent.getIntExtra(defs.BluetoothDevice.EXTRA_BOND_STATE, -1)
            if bond_state == -1:
                loop.call_soon_threadsafe(
                    bondedFuture.set_exception,
                    BleakError(f"Unexpected bond state {bond_state}"),
                )
            elif bond_state == defs.BluetoothDevice.BOND_NONE:
                loop.call_soon_threadsafe(
                    bondedFuture.set_exception,
                    BleakError(
                        f"Device with address {self.address} could not be paired with."
                    ),
                )
            elif bond_state == defs.BluetoothDevice.BOND_BONDED:
                loop.call_soon_threadsafe(bondedFuture.set_result, True)

        receiver = BroadcastReceiver(
            handleBondStateChanged,
            actions=[defs.BluetoothDevice.ACTION_BOND_STATE_CHANGED],
        )
        receiver.start()
        try:
            # See if it is already paired.
            bond_state = self.__device.getBondState()
            if bond_state == defs.BluetoothDevice.BOND_BONDED:
                return
            elif bond_state == defs.BluetoothDevice.BOND_NONE:
                logger.debug(f"Pairing to BLE device @ {self.address}")
                if not self.__device.createBond():
                    raise BleakError(
                        f"Could not initiate bonding with device @ {self.address}"
                    )
            await bondedFuture
        finally:
            await receiver.stop()

    @override
    async def unpair(self) -> None:
        """Unpair with the peripheral."""
        warnings.warn(
            "Unpairing is seemingly unavailable in the Android API at the moment."
        )

    @property
    @override
    def is_connected(self) -> bool:
        """Check connection status between this client and the server.

        Returns:
            Boolean representing connection status.

        """
        return (
            self.__callbacks is not None
            and self.__callbacks.states["onConnectionStateChange"][1]
            == defs.BluetoothProfile.STATE_CONNECTED
        )

    @property
    @override
    def mtu_size(self) -> int:
        return self.__mtu

    # GATT services methods

    async def _get_services(self) -> BleakGATTServiceCollection:
        """Get all services registered for this GATT server.

        Returns:
           A :py:class:`bleak.backends.service.BleakGATTServiceCollection` with this device's services tree.

        """
        if self.services is not None:
            return self.services

        services = BleakGATTServiceCollection()

        logger.debug("Get Services...")
        for java_service in self.__gatt.getServices():
            if (
                self._requested_services is not None
                and java_service.getUuid() not in self._requested_services
            ):
                continue

            service = BleakGATTService(
                java_service,
                java_service.getInstanceId(),
                java_service.getUuid().toString(),
            )
            services.add_service(service)

            for java_characteristic in java_service.getCharacteristics():

                characteristic = BleakGATTCharacteristic(
                    java_characteristic,
                    java_characteristic.getInstanceId(),
                    java_characteristic.getUuid().toString(),
                    gatt_char_props_to_strs((java_characteristic.getProperties())),
                    lambda: self.__mtu - 3,
                    service,
                )
                services.add_characteristic(characteristic)

                for descriptor_index, java_descriptor in enumerate(
                    java_characteristic.getDescriptors()
                ):

                    descriptor = BleakGATTDescriptor(
                        java_descriptor,
                        characteristic.handle + 1 + descriptor_index,
                        java_descriptor.getUuid().toString(),
                        characteristic,
                    )
                    services.add_descriptor(descriptor)

        self.services = services
        return self.services

    # IO methods

    @override
    async def read_gatt_char(
        self, characteristic: BleakGATTCharacteristic, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to read from.

        Returns:
            (bytearray) The read data.

        """

        (value,) = await self.__callbacks.perform_and_wait(
            dispatchApi=self.__gatt.readCharacteristic,
            dispatchParams=(characteristic.obj,),
            resultApi=("onCharacteristicRead", characteristic.handle),
        )
        value = bytearray(value)
        logger.debug(
            f"Read Characteristic {characteristic.uuid} | {characteristic.handle}: {value}"
        )
        return value

    @override
    async def read_gatt_descriptor(
        self, descriptor: BleakGATTDescriptor, **kwargs: Any
    ) -> bytearray:
        """Perform read operation on the specified GATT descriptor.

        Args:
            descriptor: The descriptor to read from.

        Returns:
            The read data.
        """
        (value,) = await self.__callbacks.perform_and_wait(
            dispatchApi=self.__gatt.readDescriptor,
            dispatchParams=(descriptor.obj,),
            resultApi=("onDescriptorRead", descriptor.uuid),
        )
        value = bytearray(value)

        logger.debug(
            f"Read Descriptor {descriptor.uuid} | {descriptor.handle}: {value}"
        )

        return value

    @override
    async def write_gatt_char(
        self, characteristic: BleakGATTCharacteristic, data: bytearray, response: bool
    ) -> None:
        if response:
            characteristic.obj.setWriteType(
                defs.BluetoothGattCharacteristic.WRITE_TYPE_DEFAULT
            )
        else:
            characteristic.obj.setWriteType(
                defs.BluetoothGattCharacteristic.WRITE_TYPE_NO_RESPONSE
            )

        characteristic.obj.setValue(data)

        await self.__callbacks.perform_and_wait(
            dispatchApi=self.__gatt.writeCharacteristic,
            dispatchParams=(characteristic.obj,),
            resultApi=("onCharacteristicWrite", characteristic.handle),
        )

        logger.debug(
            f"Write Characteristic {characteristic.uuid} | {characteristic.handle}: {data}"
        )

    @override
    async def write_gatt_descriptor(
        self,
        desc_specifier: Union[BleakGATTDescriptor, str, uuid.UUID],
        data: bytearray,
    ) -> None:
        """Perform a write operation on the specified GATT descriptor.

        Args:
            desc_specifier (BleakGATTDescriptor, str or UUID): The descriptor to write
                to, specified by either UUID or directly by the
                BleakGATTDescriptor object representing it.
            data (bytes or bytearray): The data to send.

        """
        if not isinstance(desc_specifier, BleakGATTDescriptor):
            descriptor = self.services.get_descriptor(desc_specifier)
        else:
            descriptor = desc_specifier

        if not descriptor:
            raise BleakError(f"Descriptor {desc_specifier} was not found!")

        descriptor.obj.setValue(data)

        await self.__callbacks.perform_and_wait(
            dispatchApi=self.__gatt.writeDescriptor,
            dispatchParams=(descriptor.obj,),
            resultApi=("onDescriptorWrite", descriptor.uuid),
        )

        logger.debug(
            f"Write Descriptor {descriptor.uuid} | {descriptor.handle}: {data}"
        )

    @override
    async def start_notify(
        self,
        characteristic: BleakGATTCharacteristic,
        callback: NotifyCallback,
        **kwargs,
    ) -> None:
        """
        Activate notifications/indications on a characteristic.
        """
        self._subscriptions[characteristic.handle] = callback

        assert self.__gatt is not None

        if not self.__gatt.setCharacteristicNotification(characteristic.obj, True):
            raise BleakError(
                f"Failed to enable notification for characteristic {characteristic.uuid}"
            )

        await self.write_gatt_descriptor(
            characteristic.notification_descriptor,
            defs.BluetoothGattDescriptor.ENABLE_NOTIFICATION_VALUE,
        )

    @override
    async def stop_notify(self, characteristic: BleakGATTCharacteristic) -> None:
        """Deactivate notification/indication on a specified characteristic.

        Args:
            characteristic (BleakGATTCharacteristic): The characteristic to deactivate
                notification/indication on,.

        """
        await self.write_gatt_descriptor(
            characteristic.notification_descriptor,
            defs.BluetoothGattDescriptor.DISABLE_NOTIFICATION_VALUE,
        )

        if not self.__gatt.setCharacteristicNotification(characteristic.obj, False):
            raise BleakError(
                f"Failed to disable notification for characteristic {characteristic.uuid}"
            )
        del self._subscriptions[characteristic.handle]


class _PythonBluetoothGattCallback(utils.AsyncJavaCallbacks):
    __javainterfaces__ = [
        "com.github.hbldh.bleak.PythonBluetoothGattCallback$Interface"
    ]

    def __init__(self, client, loop):
        super().__init__(loop)
        self._client = client
        self.java = defs.PythonBluetoothGattCallback(self)

    def result_state(self, status, resultApi, *data):
        if status == defs.BluetoothGatt.GATT_SUCCESS:
            failure_str = None
        else:
            failure_str = defs.GATT_STATUS_STRINGS.get(status, status)
        self._loop.call_soon_threadsafe(
            self._result_state_unthreadsafe, failure_str, resultApi, data
        )

    @java_method("(II)V")
    def onConnectionStateChange(self, status, new_state):
        try:
            self.result_state(status, "onConnectionStateChange", new_state)
        except BleakError:
            pass
        if (
            new_state == defs.BluetoothProfile.STATE_DISCONNECTED
            and self._client._disconnected_callback is not None
        ):
            self._client._disconnected_callback()

    @java_method("(II)V")
    def onMtuChanged(self, mtu, status):
        self.result_state(status, "onMtuChanged", mtu)

    @java_method("(I)V")
    def onServicesDiscovered(self, status):
        self.result_state(status, "onServicesDiscovered")

    @java_method("(I[B)V")
    def onCharacteristicChanged(self, handle, value):
        self._loop.call_soon_threadsafe(
            self._client._subscriptions[handle], bytearray(value.tolist())
        )

    @java_method("(II[B)V")
    def onCharacteristicRead(self, handle, status, value):
        self.result_state(
            status, ("onCharacteristicRead", handle), bytes(value.tolist())
        )

    @java_method("(II)V")
    def onCharacteristicWrite(self, handle, status):
        self.result_state(status, ("onCharacteristicWrite", handle))

    @java_method("(Ljava/lang/String;I[B)V")
    def onDescriptorRead(self, uuid, status, value):
        self.result_state(status, ("onDescriptorRead", uuid), bytes(value.tolist()))

    @java_method("(Ljava/lang/String;I)V")
    def onDescriptorWrite(self, uuid, status):
        self.result_state(status, ("onDescriptorWrite", uuid))



import abc
import asyncio
import inspect
from collections.abc import Callable, Coroutine, Hashable
from typing import Any, NamedTuple, Optional

from bleak.backends import BleakBackend, get_default_backend
from bleak.backends.device import BLEDevice
from bleak.exc import BleakError

# prevent tasks from being garbage collected
_background_tasks: set[asyncio.Task[None]] = set()


class AdvertisementData(NamedTuple):
    """
    Wrapper around the advertisement data that each platform returns upon discovery
    """

    local_name: Optional[str]
    """
    The local name of the device or ``None`` if not included in advertising data.
    """

    manufacturer_data: dict[int, bytes]
    """
    Dictionary of manufacturer data in bytes from the received advertisement data or empty dict if not present.

    The keys are Bluetooth SIG assigned Company Identifiers and the values are bytes.

    https://www.bluetooth.com/specifications/assigned-numbers/company-identifiers/
    """

    service_data: dict[str, bytes]
    """
    Dictionary of service data from the received advertisement data or empty dict if not present.
    """

    service_uuids: list[str]
    """
    List of service UUIDs from the received advertisement data or empty list if not present.
    """

    tx_power: Optional[int]
    """
    TX Power Level of the remote device from the received advertising data or ``None`` if not present.

    .. versionadded:: 0.17
    """

    rssi: int
    """
    The Radio Receive Signal Strength (RSSI) in dBm.

    .. versionadded:: 0.19
    """

    platform_data: tuple[Any, ...]
    """
    Tuple of platform specific data.

    This is not a stable API. The actual values may change between releases.
    """

    def __repr__(self) -> str:
        kwargs: list[str] = []
        if self.local_name:
            kwargs.append(f"local_name={repr(self.local_name)}")
        if self.manufacturer_data:
            kwargs.append(f"manufacturer_data={repr(self.manufacturer_data)}")
        if self.service_data:
            kwargs.append(f"service_data={repr(self.service_data)}")
        if self.service_uuids:
            kwargs.append(f"service_uuids={repr(self.service_uuids)}")
        if self.tx_power is not None:
            kwargs.append(f"tx_power={repr(self.tx_power)}")
        kwargs.append(f"rssi={repr(self.rssi)}")
        return f"AdvertisementData({', '.join(kwargs)})"


AdvertisementDataCallback = Callable[
    [BLEDevice, AdvertisementData],
    Optional[Coroutine[Any, Any, None]],
]
"""
Type alias for callback called when advertisement data is received.
"""

AdvertisementDataFilter = Callable[
    [BLEDevice, AdvertisementData],
    bool,
]
"""
Type alias for an advertisement data filter function.

Implementations should return ``True`` for matches, otherwise ``False``.
"""


class BaseBleakScanner(abc.ABC):
    """
    Interface for Bleak Bluetooth LE Scanners

    Args:
        detection_callback:
            Optional function that will be called each time a device is
            discovered or advertising data has changed.
        service_uuids:
            Optional list of service UUIDs to filter on. Only advertisements
            containing this advertising data will be received.
    """

    seen_devices: dict[str, tuple[BLEDevice, AdvertisementData]]
    """
    Map of device identifier to BLEDevice and most recent advertisement data.

    The key is a backend-specific identifier for the device.

    This map must be cleared when scanning starts.
    """

    def __init__(
        self,
        detection_callback: Optional[AdvertisementDataCallback],
        service_uuids: Optional[list[str]],
    ):
        super(BaseBleakScanner, self).__init__()

        self._ad_callbacks: dict[
            Hashable, Callable[[BLEDevice, AdvertisementData], None]
        ] = {}
        """
        List of callbacks to call when an advertisement is received.
        """

        if detection_callback is not None:
            self.register_detection_callback(detection_callback)

        self._service_uuids: Optional[list[str]] = (
            [u.lower() for u in service_uuids] if service_uuids is not None else None
        )

        self.seen_devices = {}

    def register_detection_callback(
        self, callback: Optional[AdvertisementDataCallback]
    ) -> Callable[[], None]:
        """
        Register a callback that is called when an advertisement event from the
        OS is received.

        The ``callback`` is a function or coroutine that takes two arguments: :class:`BLEDevice`
        and :class:`AdvertisementData`.

        Args:
            callback: A function, coroutine or ``None``.

        Returns:
            A method that can be called to unregister the callback.
        """
        error_text = "callback must be callable with 2 parameters"

        if not callable(callback):
            raise TypeError(error_text)

        handler_signature = inspect.signature(callback)

        if len(handler_signature.parameters) != 2:
            raise TypeError(error_text)

        if inspect.iscoroutinefunction(callback):

            def detection_callback(s: BLEDevice, d: AdvertisementData) -> None:
                task = asyncio.create_task(callback(s, d))
                _background_tasks.add(task)
                task.add_done_callback(_background_tasks.discard)

        else:
            detection_callback = callback

        token = object()

        self._ad_callbacks[token] = detection_callback

        def remove() -> None:
            self._ad_callbacks.pop(token, None)

        return remove

    def is_allowed_uuid(self, service_uuids: Optional[list[str]]) -> bool:
        """
        Check if the advertisement data contains any of the service UUIDs
        matching the filter. If no filter is set, this will always return
        ``True``.

        Args:
            service_uuids: The service UUIDs from the advertisement data.

        Returns:
            ``True`` if the advertisement data should be allowed or ``False``
             if the advertisement data should be filtered out.
        """
        # Backends will make best effort to filter out advertisements that
        # don't match the service UUIDs, but if other apps are scanning at the
        # same time or something like that, we may still receive advertisements
        # that don't match. So we need to do more filtering here to get the
        # expected behavior.

        if not self._service_uuids:
            # if there is no filter, everything is allowed
            return True

        if not service_uuids:
            # if there is a filter the advertisement data doesn't contain any
            # service UUIDs, filter it out
            return False

        for uuid in service_uuids:
            if uuid in self._service_uuids:
                # match was found, keep this advertisement
                return True

        # there were no matching service uuids, filter this one out
        return False

    def call_detection_callbacks(
        self, device: BLEDevice, advertisement_data: AdvertisementData
    ) -> None:
        """
        Calls all registered detection callbacks.

        Backend implementations should call this method when an advertisement
        event is received from the OS.
        """

        for callback in self._ad_callbacks.values():
            callback(device, advertisement_data)

    def create_or_update_device(
        self,
        key: str,
        address: str,
        name: Optional[str],
        details: Any,
        adv: AdvertisementData,
    ) -> BLEDevice:
        """
        Creates or updates a device in :attr:`seen_devices`.

        Args:
            key: A backend-specific identifier for the device.
            address: The Bluetooth address of the device (UUID on macOS).
            name: The OS display name for the device.
            details: The platform-specific handle for the device.
            adv: The most recent advertisement data received.

        Returns:
            The updated device.
        """

        try:
            device, _ = self.seen_devices[key]

            device.name = name
        except KeyError:
            device = BLEDevice(address, name, details)

        self.seen_devices[key] = (device, adv)

        return device

    @abc.abstractmethod
    async def start(self) -> None:
        """Start scanning for devices"""
        raise NotImplementedError()

    @abc.abstractmethod
    async def stop(self) -> None:
        """Stop scanning for devices"""
        raise NotImplementedError()


def get_platform_scanner_backend_type() -> tuple[type[BaseBleakScanner], BleakBackend]:
    """
    Gets the platform-specific :class:`BaseBleakScanner` type.
    """
    backend = get_default_backend()
    match backend:
        case BleakBackend.P4ANDROID:
            from bleak.backends.p4android.scanner import BleakScannerP4Android

            return (BleakScannerP4Android, backend)

        case BleakBackend.BLUEZ_DBUS:
            from bleak.backends.bluezdbus.scanner import BleakScannerBlueZDBus

            return (BleakScannerBlueZDBus, backend)

        case BleakBackend.PYTHONISTA_CB:
            try:
                from bleak_pythonista import BleakScannerPythonistaCB

                return (BleakScannerPythonistaCB, backend)
            except ImportError as e:
                raise ImportError(
                    "Ensure you have `bleak-pythonista` package installed."
                ) from e

        case BleakBackend.CORE_BLUETOOTH:
            from bleak.backends.corebluetooth.scanner import BleakScannerCoreBluetooth

            return (BleakScannerCoreBluetooth, backend)

        case BleakBackend.WIN_RT:
            from bleak.backends.winrt.scanner import BleakScannerWinRT

            return (BleakScannerWinRT, backend)

        case _:
            raise BleakError(f"Unsupported backend: {backend}")
